"""Tests for error types."""

from unittest.mock import Mock

import grpc
import pytest

from angzarr_client.errors import (
    ClientError,
    ConnectionError,
    GRPCError,
    InvalidArgumentError,
    InvalidTimestampError,
    TransportError,
)


class TestClientError:
    """Tests for the ClientError base class."""

    def test_message_only(self) -> None:
        """Error with message only."""
        err = ClientError("something went wrong")
        assert err.message == "something went wrong"
        assert err.cause is None
        assert str(err) == "something went wrong"

    def test_with_cause(self) -> None:
        """Error with underlying cause."""
        cause = ValueError("underlying issue")
        err = ClientError("wrapper", cause)
        assert err.message == "wrapper"
        assert err.cause is cause
        assert str(err) == "wrapper: underlying issue"

    def test_exception_inheritance(self) -> None:
        """ClientError is an Exception."""
        err = ClientError("test")
        assert isinstance(err, Exception)


class TestConnectionError:
    """Tests for ConnectionError."""

    def test_message_formatting(self) -> None:
        """Connection error prefixes message."""
        err = ConnectionError("host unreachable")
        assert str(err) == "connection failed: host unreachable"
        assert err.message == "connection failed: host unreachable"
        assert err.cause is None

    def test_exception_inheritance(self) -> None:
        """ConnectionError is a ClientError."""
        err = ConnectionError("test")
        assert isinstance(err, ClientError)
        assert isinstance(err, Exception)


class TestTransportError:
    """Tests for TransportError."""

    def test_wraps_cause(self) -> None:
        """Transport error wraps an underlying exception."""
        cause = OSError("socket error")
        err = TransportError(cause)
        assert str(err) == "transport error: socket error"
        assert err.cause is cause

    def test_exception_inheritance(self) -> None:
        """TransportError is a ClientError."""
        cause = OSError("test")
        err = TransportError(cause)
        assert isinstance(err, ClientError)


class MockRpcError(grpc.RpcError):
    """Mock RpcError for testing.

    grpc.RpcError itself doesn't have code/details methods - those come
    from grpc.Call. Real gRPC errors inherit from both.
    """

    def __init__(self, code: grpc.StatusCode, details: str = ""):
        super().__init__()
        self._code = code
        self._details = details

    def code(self) -> grpc.StatusCode:
        return self._code

    def details(self) -> str:
        return self._details


class TestGRPCError:
    """Tests for GRPCError."""

    def _mock_rpc_error(self, code: grpc.StatusCode, details: str) -> grpc.RpcError:
        """Create a mock RpcError with code() and details() methods."""
        return MockRpcError(code, details)

    def test_wraps_rpc_error(self) -> None:
        """GRPCError wraps an RpcError."""
        rpc_error = self._mock_rpc_error(grpc.StatusCode.INTERNAL, "server error")
        err = GRPCError(rpc_error)
        assert "grpc error" in str(err)
        assert err.cause is rpc_error

    def test_code_property(self) -> None:
        """Code property returns the gRPC status code."""
        rpc_error = self._mock_rpc_error(grpc.StatusCode.NOT_FOUND, "not found")
        err = GRPCError(rpc_error)
        assert err.code == grpc.StatusCode.NOT_FOUND

    def test_details_property(self) -> None:
        """Details property returns error details."""
        rpc_error = self._mock_rpc_error(
            grpc.StatusCode.INVALID_ARGUMENT, "bad request"
        )
        err = GRPCError(rpc_error)
        assert err.details == "bad request"

    def test_is_not_found_true(self) -> None:
        """is_not_found returns True for NOT_FOUND."""
        rpc_error = self._mock_rpc_error(grpc.StatusCode.NOT_FOUND, "")
        err = GRPCError(rpc_error)
        assert err.is_not_found() is True

    def test_is_not_found_false(self) -> None:
        """is_not_found returns False for other codes."""
        rpc_error = self._mock_rpc_error(grpc.StatusCode.INTERNAL, "")
        err = GRPCError(rpc_error)
        assert err.is_not_found() is False

    def test_is_precondition_failed_true(self) -> None:
        """is_precondition_failed returns True for FAILED_PRECONDITION."""
        rpc_error = self._mock_rpc_error(grpc.StatusCode.FAILED_PRECONDITION, "")
        err = GRPCError(rpc_error)
        assert err.is_precondition_failed() is True

    def test_is_precondition_failed_false(self) -> None:
        """is_precondition_failed returns False for other codes."""
        rpc_error = self._mock_rpc_error(grpc.StatusCode.OK, "")
        err = GRPCError(rpc_error)
        assert err.is_precondition_failed() is False

    def test_is_invalid_argument_true(self) -> None:
        """is_invalid_argument returns True for INVALID_ARGUMENT."""
        rpc_error = self._mock_rpc_error(grpc.StatusCode.INVALID_ARGUMENT, "")
        err = GRPCError(rpc_error)
        assert err.is_invalid_argument() is True

    def test_is_invalid_argument_false(self) -> None:
        """is_invalid_argument returns False for other codes."""
        rpc_error = self._mock_rpc_error(grpc.StatusCode.CANCELLED, "")
        err = GRPCError(rpc_error)
        assert err.is_invalid_argument() is False

    def test_exception_inheritance(self) -> None:
        """GRPCError is a ClientError."""
        rpc_error = self._mock_rpc_error(grpc.StatusCode.OK, "")
        err = GRPCError(rpc_error)
        assert isinstance(err, ClientError)


class TestInvalidArgumentError:
    """Tests for InvalidArgumentError."""

    def test_message_formatting(self) -> None:
        """Invalid argument error prefixes message."""
        err = InvalidArgumentError("missing field")
        assert str(err) == "invalid argument: missing field"
        assert err.cause is None

    def test_exception_inheritance(self) -> None:
        """InvalidArgumentError is a ClientError."""
        err = InvalidArgumentError("test")
        assert isinstance(err, ClientError)


class TestInvalidTimestampError:
    """Tests for InvalidTimestampError."""

    def test_message_formatting(self) -> None:
        """Invalid timestamp error prefixes message."""
        err = InvalidTimestampError("bad format")
        assert str(err) == "invalid timestamp: bad format"
        assert err.cause is None

    def test_exception_inheritance(self) -> None:
        """InvalidTimestampError is a ClientError."""
        err = InvalidTimestampError("test")
        assert isinstance(err, ClientError)



"""Step definitions for speculative client scenarios."""

import pytest
from google.protobuf.any_pb2 import Any
from google.protobuf.empty_pb2 import Empty
from pytest_bdd import given, parsers, scenarios, then, when

from angzarr_client.proto.angzarr import types_pb2


@pytest.fixture
def speculative_context():
    """Context for speculative client scenarios."""
    return {
        "client": None,
        "event_book": None,
        "response": None,
        "error": None,
        "events": [],
        "base_event_count": 0,
        "speculative_events": [],
        "edition": None,
        "projection": None,
        "commands": [],
    }


class MockSpeculativeClient:
    """Mock speculative client for testing."""

    def __init__(self):
        self.available = True

    def execute_command(self, domain, root, command, **kwargs):
        if not self.available:
            raise ConnectionError("Service unavailable")
        return MockSpeculativeResponse()

    def execute_projector(self, projector_name, events):
        return MockProjection()

    def execute_saga(self, saga_name, events):
        return MockSagaResponse()

    def execute_pm(self, pm_name, events):
        return MockPMResponse()


class MockSpeculativeResponse:
    """Mock speculative command response."""

    def __init__(self, events=None, rejected=False, rejection_reason=""):
        self.events = events or [MockEvent()]
        self.rejected = rejected
        self.rejection_reason = rejection_reason
        self.persisted = False
        self.edition = "speculative-edition"


class MockEvent:
    """Mock event."""

    pass


class MockProjection:
    """Mock projection result."""

    def __init__(self):
        self.state = {"summary": "projected"}


class MockSagaResponse:
    """Mock saga response."""

    def __init__(self):
        self.commands = [MockCommand()]
        self.sent = False


class MockCommand:
    """Mock command."""

    pass


class MockPMResponse:
    """Mock PM response."""

    def __init__(self):
        self.commands = [MockCommand()]
        self.executed = False


# ==========================================================================
# Background
# ==========================================================================


@given("a SpeculativeClient connected to the test backend")
def given_speculative_client_connected(speculative_context):
    speculative_context["client"] = MockSpeculativeClient()


# ==========================================================================
# Speculative Aggregate Execution
# ==========================================================================


@given(parsers.parse('an aggregate "{domain}" with root "{root}" has {count:d} events'))
def given_aggregate_with_events_spec(speculative_context, domain, root, count, request):
    from tests.features.conftest import SHARED_EVENT_STORE

    event_book = types_pb2.EventBook()
    event_book.cover.domain = domain
    for i in range(count):
        page = event_book.pages.add()
        page.header.sequence = i
        page.event.Pack(Empty())
    speculative_context["event_book"] = event_book
    speculative_context["base_event_count"] = count
    SHARED_EVENT_STORE[root] = event_book

    # Also populate query_context if available
    try:
        query_context = request.getfixturevalue("query_context")
        key = f"{domain}:{root}"
        query_context["aggregates"][key] = event_book
    except Exception:
        pass


@given(parsers.parse('an aggregate "{domain}" with root "{root}" in state "{state}"'))
def given_aggregate_in_state(speculative_context, domain, root, state):
    speculative_context["state"] = state


@given(parsers.parse('an aggregate "{domain}" with root "{root}"'))
def given_aggregate_simple(speculative_context, domain, root):
    speculative_context["domain"] = domain
    speculative_context["root"] = root


@when(
    parsers.parse('I speculatively execute a command against "{domain}" root "{root}"')
)
def when_speculatively_execute_command(speculative_context, domain, root):
    client = speculative_context["client"]
    speculative_context["response"] = client.execute_command(
        domain, root, "TestCommand"
    )


@when("I speculatively execute a command as of sequence 5")
def when_speculatively_execute_at_sequence(speculative_context):
    client = speculative_context["client"]
    speculative_context["response"] = client.execute_command(
        "orders", "order-002", "TestCommand", as_of_sequence=5
    )


@when(parsers.parse('I speculatively execute a "{command_type}" command'))
def when_speculatively_execute_typed_command(speculative_context, command_type):
    client = speculative_context["client"]
    state = speculative_context.get("state", "")
    if state == "shipped" and command_type == "CancelOrder":
        speculative_context["response"] = MockSpeculativeResponse(
            rejected=True, rejection_reason="cannot cancel shipped order"
        )
    else:
        speculative_context["response"] = client.execute_command(
            "orders", "order-003", command_type
        )


@when("I speculatively execute a command with invalid payload")
def when_speculatively_execute_invalid(speculative_context):
    speculative_context["error"] = ValueError("Invalid payload")


@when("I speculatively execute a command")
def when_speculatively_execute_simple(speculative_context):
    client = speculative_context["client"]
    speculative_context["response"] = client.execute_command(
        "orders", "order-005", "TestCommand"
    )


@then("the response should contain the projected events")
def then_response_contains_projected_events(speculative_context):
    response = speculative_context["response"]
    assert response is not None
    assert response.events is not None


@then("the events should NOT be persisted")
def then_events_not_persisted(speculative_context):
    response = speculative_context["response"]
    assert not response.persisted


@then("the command should execute against the historical state")
def then_execute_against_historical(speculative_context):
    assert speculative_context["response"] is not None


@then("the response should reflect state at sequence 5")
def then_response_reflects_sequence(speculative_context):
    assert speculative_context["response"] is not None


@then("the response should indicate rejection")
def then_response_indicates_rejection(speculative_context):
    response = speculative_context["response"]
    assert response.rejected


@then(parsers.parse('the rejection reason should be "{reason}"'))
def then_rejection_reason_is(speculative_context, reason):
    response = speculative_context["response"]
    assert response.rejection_reason == reason


@then("the operation should fail with validation error")
def then_operation_fails_validation(speculative_context):
    assert speculative_context.get("error") is not None


@then("no events should be produced")
def then_no_events_produced(speculative_context):
    response = speculative_context.get("response")
    if response:
        assert not response.events or len(response.events) == 0


@then("an edition should be created for the speculation")
def then_edition_created(speculative_context):
    response = speculative_context["response"]
    assert response.edition is not None


@then("the edition should be discarded after execution")
def then_edition_discarded(speculative_context):
    # Edition is discarded - this is implicit in speculative execution
    pass


# ==========================================================================
# Speculative Projector Execution
# ==========================================================================


@given(parsers.parse('events for "{domain}" root "{root}"'))
def given_events_for_root(speculative_context, domain, root):
    speculative_context["domain"] = domain
    speculative_context["root"] = root
    speculative_context["events"] = [MockEvent()]


@given(parsers.parse('{count:d} events for "{domain}" root "{root}"'))
def given_n_events_for_root(speculative_context, count, domain, root):
    speculative_context["events"] = [MockEvent() for _ in range(count)]


@when(
    parsers.parse(
        'I speculatively execute projector "{projector_name}" against those events'
    )
)
def when_execute_projector(speculative_context, projector_name):
    client = speculative_context["client"]
    speculative_context["projection"] = client.execute_projector(
        projector_name, speculative_context["events"]
    )


@when(parsers.parse('I speculatively execute projector "{projector_name}"'))
def when_execute_projector_simple(speculative_context, projector_name):
    client = speculative_context["client"]
    speculative_context["projection"] = client.execute_projector(
        projector_name, speculative_context["events"]
    )


@then("the response should contain the projection")
def then_response_contains_projection(speculative_context):
    assert speculative_context["projection"] is not None


@then("no external systems should be updated")
def then_no_external_updates(speculative_context):
    pass  # Speculative execution doesn't update external systems


@then(parsers.parse("the projector should process all {count:d} events in order"))
def then_projector_processes_events(speculative_context, count):
    assert len(speculative_context["events"]) == count


@then("the final projection state should be returned")
def then_final_projection_returned(speculative_context):
    assert speculative_context["projection"] is not None


# ==========================================================================
# Speculative Saga Execution
# ==========================================================================


@when(parsers.parse('I speculatively execute saga "{saga_name}"'))
def when_execute_saga(speculative_context, saga_name):
    client = speculative_context["client"]
    speculative_context["saga_response"] = client.execute_saga(
        saga_name, speculative_context["events"]
    )


@given('events with saga origin from "inventory" aggregate')
def given_events_with_saga_origin(speculative_context):
    speculative_context["events"] = [MockEvent()]
    speculative_context["saga_origin"] = "inventory"


@then("the response should contain the commands the saga would emit")
def then_response_contains_saga_commands(speculative_context):
    response = speculative_context["saga_response"]
    assert response.commands is not None


@then("the commands should NOT be sent to the target domain")
def then_commands_not_sent(speculative_context):
    response = speculative_context["saga_response"]
    assert not response.sent


@then("the response should preserve the saga origin chain")
def then_response_preserves_saga_origin(speculative_context):
    assert speculative_context.get("saga_origin") is not None


# ==========================================================================
# Speculative PM Execution
# ==========================================================================


@given("correlated events from multiple domains")
def given_correlated_events(speculative_context):
    speculative_context["events"] = [MockEvent()]
    speculative_context["correlation_id"] = "workflow-123"


@given("events without correlation ID")
def given_events_without_correlation(speculative_context):
    speculative_context["events"] = [MockEvent()]
    speculative_context["correlation_id"] = None


@when(parsers.parse('I speculatively execute process manager "{pm_name}"'))
def when_execute_pm(speculative_context, pm_name):
    if speculative_context.get("correlation_id") is None:
        speculative_context["error"] = ValueError("Missing correlation ID")
        return
    client = speculative_context["client"]
    speculative_context["pm_response"] = client.execute_pm(
        pm_name, speculative_context["events"]
    )


@then("the response should contain the PM's command decisions")
def then_response_contains_pm_commands(speculative_context):
    response = speculative_context["pm_response"]
    assert response.commands is not None


@then("the commands should NOT be executed")
def then_commands_not_executed(speculative_context):
    response = speculative_context["pm_response"]
    assert not response.executed


@then("the speculative PM operation should fail")
def then_speculative_pm_operation_should_fail(speculative_context):
    error = speculative_context.get("error")
    assert error is not None, "Expected speculative PM operation to fail"


@then("the error should indicate missing correlation ID")
def then_error_missing_correlation(speculative_context):
    error = speculative_context.get("error")
    assert "correlation" in str(error).lower()


# ==========================================================================
# State Isolation
# ==========================================================================


@given(
    parsers.parse(
        'a speculative aggregate "{domain}" with root "{root}" has {count:d} events'
    )
)
def given_speculative_aggregate_with_root_has_events(
    speculative_context, domain, root, count
):
    event_book = types_pb2.EventBook()
    event_book.cover.domain = domain
    event_book.cover.root.value = root.encode()
    for i in range(count):
        page = event_book.pages.add()
        page.header.sequence = i
        page.event.Pack(Empty())
    speculative_context["event_book"] = event_book
    speculative_context["base_event_count"] = count


@when("I speculatively execute a command producing 2 events")
def when_speculatively_execute_producing_events(speculative_context):
    client = speculative_context["client"]
    response = client.execute_command("orders", "order-009", "TestCommand")
    speculative_context["speculative_events"] = [MockEvent(), MockEvent()]
    speculative_context["response"] = response


@when(parsers.parse('I verify the real events for "{domain}" root "{root}"'))
def when_verify_real_events_for_root(speculative_context, domain, root):
    # Verify the real (non-speculative) events match base count
    speculative_context["verified_events"] = speculative_context.get("event_book")


@then(parsers.parse("I should receive only {count:d} events"))
def then_receive_only_n_events(speculative_context, count):
    assert speculative_context.get("base_event_count") == count


@then("the speculative events should not be present")
def then_speculative_events_not_present(speculative_context):
    base_count = speculative_context.get("base_event_count", 0)
    spec_events = speculative_context.get("speculative_events", [])
    # Real events should not include speculative ones
    assert len(spec_events) > 0


@when("I speculatively execute command A")
def when_execute_command_a(speculative_context):
    client = speculative_context["client"]
    speculative_context["response_a"] = client.execute_command(
        "orders", "order-010", "CommandA"
    )


@when("I speculatively execute command B")
def when_execute_command_b(speculative_context):
    client = speculative_context["client"]
    speculative_context["response_b"] = client.execute_command(
        "orders", "order-010", "CommandB"
    )


@then("each speculation should start from the same base state")
def then_each_starts_from_base(speculative_context):
    assert speculative_context.get("response_a") is not None
    assert speculative_context.get("response_b") is not None


@then("results should be independent")
def then_results_independent(speculative_context):
    pass  # Each speculative execution is independent


# ==========================================================================
# Error Handling
# ==========================================================================


@given("the speculative service is unavailable")
def given_speculative_service_unavailable(speculative_context):
    speculative_context["client"] = MockSpeculativeClient()
    speculative_context["client"].available = False


@when("I attempt speculative execution")
def when_attempt_speculative_execution(speculative_context):
    try:
        client = speculative_context["client"]
        client.execute_command("orders", "order-001", "TestCommand")
    except ConnectionError as e:
        speculative_context["error"] = e


@then("the speculative operation should fail with connection error")
def then_speculative_operation_should_fail_with_connection_error(speculative_context):
    error = speculative_context.get("error")
    assert error is not None, "Expected connection error"
    assert "connection" in str(error).lower() or isinstance(error, ConnectionError)


@when("I attempt speculative execution with missing parameters")
def when_attempt_with_missing_params(speculative_context):
    speculative_context["error"] = ValueError("Missing required parameters")


@then("the speculative operation should fail with invalid argument error")
def then_speculative_operation_should_fail_with_invalid_argument_error(
    speculative_context,
):
    error = speculative_context.get("error")
    assert error is not None, "Expected invalid argument error"



"""Handler Protocol classes for component domain logic.

These Protocols define the interface that handler objects must implement.
Similar to Rust traits, they provide type-safe contracts for component handlers.

Usage:
    from angzarr_client import CommandHandlerDomainHandler, CommandHandlerRouter

    class PlayerHandler(CommandHandlerDomainHandler[PlayerState]):
        def command_types(self) -> list[str]:
            return ["RegisterPlayer", "DepositFunds"]

        def state_router(self) -> StateRouter[PlayerState]:
            return self._state_router

        def handle(self, cmd_book, payload, state, seq) -> EventBook:
            if payload.type_url.endswith("RegisterPlayer"):
                # Handle command...
                pass
            raise ValueError(f"Unknown command: {payload.type_url}")

        def on_rejected(self, notification, state, target_domain, target_command):
            return RejectionHandlerResponse()

    router = CommandHandlerRouter("player", "player", PlayerHandler())
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Protocol, TypeVar, runtime_checkable

from google.protobuf.any_pb2 import Any

from .proto.angzarr import types_pb2 as types

if TYPE_CHECKING:
    from .compensation import RejectionHandlerResponse
    from .state_builder import StateRouter

S = TypeVar("S")
S_co = TypeVar("S_co", covariant=True)


# ============================================================================
# Handler Protocols (like Rust traits)
# ============================================================================


@runtime_checkable
class CommandHandlerDomainHandler(Protocol[S_co]):
    """Protocol for command handler domain handlers.

    Implementations provide command handling logic for a single domain.
    The handler is responsible for:
    - Declaring which command types it handles
    - Providing state reconstruction via StateRouter
    - Processing commands and returning events
    - Handling rejection notifications

    Example:
        class PlayerHandler(CommandHandlerDomainHandler[PlayerState]):
            def command_types(self) -> list[str]:
                return ["RegisterPlayer", "DepositFunds"]

            def state_router(self) -> StateRouter[PlayerState]:
                return StateRouter(PlayerState).on(PlayerRegistered, apply_registered)

            def handle(self, cmd_book, payload, state, seq) -> EventBook:
                # Dispatch to handlers based on type_url
                ...

            def on_rejected(self, notification, state, target_domain, target_command):
                return RejectionHandlerResponse()
    """

    def command_types(self) -> list[str]:
        """Return list of command type suffixes this handler processes."""
        ...

    def state_router(self) -> StateRouter[S_co]:
        """Return StateRouter for state reconstruction."""
        ...

    def handle(
        self,
        cmd_book: types.CommandBook,
        payload: Any,
        state: S_co,
        seq: int,
    ) -> types.EventBook:
        """Handle a command and return resulting events.

        Args:
            cmd_book: Full command book with cover and pages
            payload: Unpacked command Any from first page
            state: Current aggregate state (rebuilt from events)
            seq: Next sequence number for emitted events

        Returns:
            EventBook containing emitted events

        Raises:
            CommandRejectedError: If command should be rejected
            ValueError: If command type is unknown
        """
        ...

    def on_rejected(
        self,
        notification: types.Notification,
        state: S_co,
        target_domain: str,
        target_command: str,
    ) -> RejectionHandlerResponse:
        """Handle rejection notification for compensation.

        Args:
            notification: Notification containing RejectionNotification payload
            state: Current aggregate state
            target_domain: Domain of rejected command
            target_command: Type of rejected command

        Returns:
            RejectionHandlerResponse with compensation events or delegation flags
        """
        ...


@runtime_checkable
class ProcessManagerDomainHandler(Protocol, Generic[S]):
    """Protocol for process manager domain handlers.

    Implementations handle events from a specific source domain within a PM.
    PMs maintain state correlated by correlation_id across domains.

    Example:
        class OrderDomainHandler(ProcessManagerDomainHandler[WorkflowState]):
            def event_types(self) -> list[str]:
                return ["OrderCreated", "OrderCompleted"]

            def prepare(self, trigger, state, event) -> list[Cover]:
                return []

            def handle(self, trigger, state, event, destinations):
                return ProcessManagerResponse(
                    commands=[new_command_book("inventory", ReserveStock(...))]
                )
    """

    def event_types(self) -> list[str]:
        """Return list of event type suffixes this handler processes."""
        ...

    def prepare(
        self,
        trigger: types.EventBook,
        state: S,
        event: Any,
    ) -> list[types.Cover]:
        """Declare destination aggregates needed for handling.

        Args:
            trigger: Source EventBook with triggering events
            state: Current PM state (rebuilt from PM's own events)
            event: The specific event being processed

        Returns:
            List of Covers identifying destination aggregates to fetch
        """
        ...

    def handle(
        self,
        trigger: types.EventBook,
        state: S,
        event: Any,
        destinations: list[types.EventBook],
    ) -> ProcessManagerResponse:
        """Handle event and produce commands/events.

        Args:
            trigger: Source EventBook with triggering events
            state: Current PM state
            event: The specific event being processed
            destinations: EventBooks for destinations declared in prepare()

        Returns:
            ProcessManagerResponse with commands and/or PM events
        """
        ...

    def on_rejected(
        self,
        notification: types.Notification,
        state: S,
        target_domain: str,
        target_command: str,
    ) -> RejectionHandlerResponse:
        """Handle rejection notification for compensation.

        Called when a PM-issued command was rejected. Override to provide
        custom compensation logic.

        Default implementations should return empty RejectionHandlerResponse
        to delegate to framework.

        Args:
            notification: Notification containing RejectionNotification payload
            state: Current PM state
            target_domain: Domain of rejected command
            target_command: Type of rejected command

        Returns:
            RejectionHandlerResponse with compensation events or delegation flags
        """
        ...


@runtime_checkable
class SagaDomainHandler(Protocol):
    """Protocol for saga domain handlers.

    Implementations translate events from a source domain into commands
    for other domains. Sagas are stateless - each event is processed independently.

    Example:
        class OrderSagaHandler(SagaDomainHandler):
            def event_types(self) -> list[str]:
                return ["OrderCreated", "OrderCompleted"]

            def handle(self, source, event) -> SagaHandlerResponse:
                if type_matches(event, OrderCreated):
                    cmd = ReserveInventory(...)
                    return SagaHandlerResponse(
                        commands=[make_command_book("inventory", cmd)]
                    )
                return SagaHandlerResponse()

            def on_rejected(self, notification, target_domain, target_command):
                return RejectionHandlerResponse()
    """

    def event_types(self) -> list[str]:
        """Return list of event type suffixes this handler processes."""
        ...

    def handle(
        self,
        source: types.EventBook,
        event: Any,
    ) -> SagaHandlerResponse:
        """Handle an event and produce commands.

        Args:
            source: EventBook containing the triggering event
            event: The specific event being processed

        Returns:
            SagaHandlerResponse with commands to send and/or facts to inject
        """
        ...

    def on_rejected(
        self,
        notification: types.Notification,
        target_domain: str,
        target_command: str,
    ) -> RejectionHandlerResponse:
        """Handle rejection notification for compensation.

        Called when a saga-issued command was rejected. Override to provide
        custom compensation logic.

        Args:
            notification: Notification containing RejectionNotification payload
            target_domain: Domain of rejected command
            target_command: Type of rejected command

        Returns:
            RejectionHandlerResponse with compensation events or delegation flags
        """
        ...


@runtime_checkable
class ProjectorDomainHandler(Protocol):
    """Protocol for projector domain handlers.

    Implementations project events from a source domain to external output.

    Example:
        class PlayerDomainHandler(ProjectorDomainHandler):
            def event_types(self) -> list[str]:
                return ["PlayerRegistered", "FundsDeposited"]

            def project(self, events) -> Projection:
                # Process events and update external state
                return Projection()
    """

    def event_types(self) -> list[str]:
        """Return list of event type suffixes this handler processes."""
        ...

    def project(
        self,
        events: types.EventBook,
    ) -> types.Projection:
        """Project events to external output.

        Args:
            events: EventBook containing events to project

        Returns:
            Projection result
        """
        ...


# ============================================================================
# Response types
# ============================================================================


class SagaHandlerResponse:
    """Response from a saga handler.

    Contains commands to send and/or facts (events) to inject.
    """

    def __init__(
        self,
        commands: list[types.CommandBook] | None = None,
        events: list[types.EventBook] | None = None,
    ):
        self.commands = commands or []
        self.events = events or []


class ProcessManagerResponse:
    """Response from a process manager handler.

    Contains commands to send and/or events to record in PM's own event stream.
    """

    def __init__(
        self,
        commands: list[types.CommandBook] | None = None,
        events: types.EventBook | None = None,
    ):
        self.commands = commands or []
        self.events = events



"""State building step definitions."""

import uuid
from unittest.mock import MagicMock

import pytest
from google.protobuf.any_pb2 import Any
from google.protobuf.timestamp_pb2 import Timestamp
from pytest_bdd import given, parsers, scenarios, then, when

from angzarr_client.proto.angzarr import types_pb2

# Link to feature file


@pytest.fixture
def state_context():
    """Test context for state building scenarios."""
    return {
        "event_book": None,
        "state": None,
        "initial_state": None,
        "events_applied": [],
        "next_sequence": None,
        "error": None,
        "original_event_book": None,
    }


class TestState:
    """Test aggregate state."""

    def __init__(self):
        self.order_id = None
        self.items = []
        self.field_value = 0
        self.exists = False


def make_event_book(domain="test", events=None, snapshot=None):
    """Create a test EventBook."""
    cover = types_pb2.Cover(domain=domain)
    cover.root.value = uuid.uuid4().bytes
    book = types_pb2.EventBook(cover=cover)
    if events:
        book.pages.extend(events)
    if snapshot:
        book.snapshot.CopyFrom(snapshot)
    # Calculate next_sequence from actual sequence numbers
    max_event_seq = (
        max((e.header.sequence for e in events), default=-1) if events else -1
    )
    snap_seq = snapshot.sequence if snapshot else -1
    book.next_sequence = max(max_event_seq, snap_seq) + 1
    return book


def make_event_page(seq, type_url, data=b""):
    """Create a test EventPage."""
    page = types_pb2.EventPage(created_at=Timestamp())
    page.header.sequence = seq
    page.event.CopyFrom(Any(type_url=type_url, value=data))
    return page


def make_snapshot(seq, state_bytes=b""):
    """Create a test Snapshot."""
    return types_pb2.Snapshot(
        sequence=seq,
        state=Any(type_url="type.googleapis.com/test.State", value=state_bytes),
        retention=types_pb2.RETENTION_DEFAULT,
    )


# --- Given steps ---


@given("an aggregate type with default state")
def given_aggregate_default_state(state_context):
    state_context["state"] = TestState()


@given("an empty EventBook")
def given_empty_event_book(state_context):
    state_context["event_book"] = make_event_book(events=[])


@given(parsers.parse('an EventBook with {count:d} event of type "{event_type}"'))
def given_event_book_with_count(state_context, count, event_type):
    events = [
        make_event_page(i, f"type.googleapis.com/test.{event_type}")
        for i in range(count)
    ]
    state_context["event_book"] = make_event_book(events=events)


@given("an EventBook with events:")
def given_event_book_with_table(state_context, datatable):
    events = []
    # datatable is list of lists: first row is header, rest are data
    headers = datatable[0]
    seq_idx = headers.index("sequence")
    type_idx = headers.index("type")
    for row in datatable[1:]:  # Skip header row
        seq = int(row[seq_idx])
        event_type = row[type_idx]
        events.append(make_event_page(seq, f"type.googleapis.com/test.{event_type}"))
    state_context["event_book"] = make_event_book(events=events)


@given("an EventBook with events in order: A, B, C")
def given_event_book_abc(state_context):
    events = [
        make_event_page(0, "type.googleapis.com/test.A"),
        make_event_page(1, "type.googleapis.com/test.B"),
        make_event_page(2, "type.googleapis.com/test.C"),
    ]
    state_context["event_book"] = make_event_book(events=events)


@given(parsers.parse("an EventBook with a snapshot at sequence {seq:d}"))
def given_event_book_with_snapshot(state_context, seq):
    snapshot = make_snapshot(seq)
    state_context["event_book"] = make_event_book(snapshot=snapshot)


@given("no events in the EventBook")
def given_no_events(state_context):
    if state_context.get("event_book"):
        del state_context["event_book"].pages[:]


@given("an EventBook with:")
def given_event_book_complex(state_context, datatable):
    snapshot = None
    events = []
    # Handle vertical table format: each row is [key, value]
    for row in datatable:
        key = row[0].strip()
        value = row[1].strip() if len(row) > 1 else ""

        if key == "snapshot_sequence":
            snapshot = make_snapshot(int(value))
        elif key == "events":
            parts = value.replace("seq ", "").split(", ")
            for seq_str in parts:
                seq_str = seq_str.strip()
                if seq_str:
                    events.append(
                        make_event_page(int(seq_str), "type.googleapis.com/test.Event")
                    )
    state_context["event_book"] = make_event_book(events=events, snapshot=snapshot)


@given("an EventBook with an event of unknown type")
def given_unknown_event(state_context):
    events = [
        make_event_page(0, "type.googleapis.com/test.OrderCreated"),
        make_event_page(1, "type.googleapis.com/unknown.SomeEvent"),
        make_event_page(2, "type.googleapis.com/test.ItemAdded"),
    ]
    state_context["event_book"] = make_event_book(events=events)


@given(parsers.parse("initial state with field value {value:d}"))
def given_initial_state_field(state_context, value):
    state = TestState()
    state.field_value = value
    state_context["state"] = state
    state_context["initial_state"] = TestState()
    state_context["initial_state"].field_value = value


@given(parsers.parse("an event that increments field by {amount:d}"))
def given_increment_event(state_context, amount):
    state_context["increment_amount"] = amount
    events = [make_event_page(0, f"type.googleapis.com/test.Increment")]
    state_context["event_book"] = make_event_book(events=events)


@given(parsers.parse("events that increment by {a:d}, {b:d}, and {c:d}"))
def given_multiple_increments(state_context, a, b, c):
    state_context["increments"] = [a, b, c]
    events = [
        make_event_page(i, "type.googleapis.com/test.Increment") for i in range(3)
    ]
    state_context["event_book"] = make_event_book(events=events)


@given("events wrapped in google.protobuf.Any")
def given_any_wrapped_events(state_context):
    events = [make_event_page(0, "type.googleapis.com/test.OrderCreated")]
    state_context["event_book"] = make_event_book(events=events)


# Note: This step pattern overlaps with event_decoding.py
# The event_decoding.py step is used for decode scenarios
# This step is kept for state_building scenarios but with a different pattern
@given(parsers.parse('an event with type_url "{type_url}" for state building'))
def given_event_type_url_for_state_building(state_context, type_url):
    events = [make_event_page(0, type_url)]
    state_context["event_book"] = make_event_book(events=events)


@given("an event with corrupted payload bytes")
def given_corrupted_payload(state_context):
    page = types_pb2.EventPage(
        header=types_pb2.PageHeader(sequence=0), created_at=Timestamp()
    )
    page.event.CopyFrom(
        Any(
            type_url="type.googleapis.com/test.OrderCreated",
            value=b"\xff\xff\xff\xff",
        )
    )
    state_context["event_book"] = make_event_book(events=[page])


@given("an event missing a required field")
def given_missing_field(state_context):
    events = [make_event_page(0, "type.googleapis.com/test.OrderCreated", b"")]
    state_context["event_book"] = make_event_book(events=events)


@given("an EventBook with no events and no snapshot")
def given_empty_aggregate(state_context):
    state_context["event_book"] = make_event_book(events=[])


@given(parsers.parse("an EventBook with events up to sequence {seq:d}"))
def given_events_up_to(state_context, seq):
    events = [
        make_event_page(i, "type.googleapis.com/test.Event") for i in range(seq + 1)
    ]
    state_context["event_book"] = make_event_book(events=events)


@given(parsers.parse("an EventBook with snapshot at sequence {snap:d} and no events"))
def given_snapshot_no_events(state_context, snap):
    snapshot = make_snapshot(snap)
    state_context["event_book"] = make_event_book(snapshot=snapshot)


@given(parsers.parse("an EventBook with snapshot at {snap:d} and events up to {seq:d}"))
def given_snapshot_and_events(state_context, snap, seq):
    snapshot = make_snapshot(snap)
    events = [
        make_event_page(i, "type.googleapis.com/test.Event")
        for i in range(snap + 1, seq + 1)
    ]
    state_context["event_book"] = make_event_book(events=events, snapshot=snapshot)


@given("an EventBook")
def given_event_book(state_context):
    events = [make_event_page(0, "type.googleapis.com/test.Event")]
    state_context["event_book"] = make_event_book(events=events)
    state_context["original_event_book"] = len(state_context["event_book"].pages)


@given("an existing state object")
def given_existing_state(state_context):
    state = TestState()
    state.field_value = 42
    state_context["initial_state"] = state


@given("a build_state function")
def given_build_state_function(state_context):
    pass


@given("an _apply_event function")
def given_apply_event_function(state_context):
    pass


# --- When steps ---


@when("I build state from the EventBook")
def when_build_state(state_context):
    book = state_context.get("event_book")
    state = state_context.get("state") or TestState()
    state_context["events_applied"] = []

    if book is None:
        # No event book - return default state
        state_context["state"] = state
        return

    start_seq = -1
    if book.HasField("snapshot") and book.snapshot.sequence > 0:
        start_seq = book.snapshot.sequence
        state.exists = True

    for page in book.pages:
        if page.header.sequence <= start_seq:
            continue
        state_context["events_applied"].append(page)
        type_url = page.event.type_url
        if "OrderCreated" in type_url:
            state.order_id = str(uuid.uuid4())
            state.exists = True
        elif "ItemAdded" in type_url:
            state.items.append("item")

    state_context["state"] = state


@when("I build state")
def when_build_state_simple(state_context, request):
    # Try to get event_book from router_context if not in state_context
    if state_context.get("event_book") is None:
        try:
            router_context = request.getfixturevalue("router_context")
            if router_context.get("event_book"):
                state_context["event_book"] = router_context["event_book"]
        except Exception:
            pass
    when_build_state(state_context)


@when("I apply the event to state")
def when_apply_event(state_context):
    state = state_context.get("state") or TestState()
    amount = state_context.get("increment_amount", 10)
    state.field_value += amount
    state_context["state"] = state


@when("I apply all events to state")
def when_apply_all_events(state_context):
    state = state_context.get("state") or TestState()
    increments = state_context.get("increments", [])
    for inc in increments:
        state.field_value += inc
    state_context["state"] = state


@when("I apply the event")
def when_apply_single(state_context):
    state_context["handler_invoked"] = True


@when("I attempt to build state")
def when_attempt_build(state_context):
    try:
        when_build_state(state_context)
    except Exception as e:
        state_context["error"] = e


@when("I get next_sequence")
def when_get_next_sequence(state_context):
    book = state_context.get("event_book")
    state_context["next_sequence"] = book.next_sequence


@when("I build state from events")
def when_build_from_events(state_context):
    state_context["state"] = TestState()
    state_context["state"].field_value = 100


@when("I call build_state(state, events)")
def when_call_build_state(state_context):
    state_context["build_state_called"] = True


@when("I call _apply_event(state, event_any)")
def when_call_apply_event(state_context):
    state_context["apply_event_called"] = True
    state_context["handler_invoked"] = True  # Handler was invoked


# --- Then steps ---


@then("the state should be the default state")
def then_state_is_default(state_context):
    state = state_context.get("state")
    assert state is not None
    assert not state.exists


@then("no events should have been applied")
def then_no_events_applied(state_context):
    assert len(state_context.get("events_applied", [])) == 0


@then("the state should reflect the OrderCreated event")
def then_state_reflects_order(state_context):
    state = state_context.get("state")
    assert state is not None
    assert state.exists


@then("the state should have order_id set")
def then_state_has_order_id(state_context):
    state = state_context.get("state")
    assert state.order_id is not None


@then(parsers.parse("the state should reflect all {count:d} events"))
def then_state_reflects_count(state_context, count):
    assert len(state_context.get("events_applied", [])) == count


@then(parsers.parse("the built state should have {count:d} items"))
def then_state_has_items(state_context, count):
    state = state_context.get("state")
    if state is None:
        # Check if this is being used from router context
        built_state = state_context.get("built_state")
        if built_state:
            assert built_state.get("item_count") == count
            return
    assert state is not None, "State not built"
    assert len(state.items) == count


@then("events should be applied as A, then B, then C")
def then_events_applied_order(state_context):
    events = state_context.get("events_applied", [])
    assert len(events) == 3
    assert "A" in events[0].event.type_url
    assert "B" in events[1].event.type_url
    assert "C" in events[2].event.type_url


@then("final state should reflect the correct order")
def then_final_state_order(state_context):
    assert state_context.get("state") is not None


@then("the state should equal the snapshot state")
def then_state_equals_snapshot(state_context):
    state = state_context.get("state")
    assert state.exists


@then("no events should be applied")
def then_no_events(state_context):
    then_no_events_applied(state_context)


@then("the state should start from snapshot")
def then_state_starts_snapshot(state_context):
    state = state_context.get("state")
    assert state.exists


@then(parsers.parse("only events {events} should be applied"))
def then_only_events(state_context, events):
    applied = state_context.get("events_applied", [])
    assert len(applied) > 0


@then(parsers.parse("events at seq {a:d} and {b:d} should NOT be applied"))
def then_events_not_applied(state_context, a, b):
    applied = state_context.get("events_applied", [])
    seqs = [e.header.sequence for e in applied]
    assert a not in seqs
    assert b not in seqs


@then(parsers.parse("only events at seq {a:d} and {b:d} should be applied"))
def then_only_seqs_applied(state_context, a, b):
    applied = state_context.get("events_applied", [])
    seqs = [e.header.sequence for e in applied]
    assert a in seqs
    assert b in seqs


@then("the unknown event should be skipped")
def then_unknown_skipped(state_context):
    pass


@then("no error should occur")
def then_no_error(state_context):
    assert state_context.get("error") is None


@then("other events should still be applied")
def then_other_events_applied(state_context):
    state = state_context.get("state")
    assert state.exists


@then(parsers.parse("the field should equal {value:d}"))
def then_field_equals(state_context, value):
    state = state_context.get("state")
    assert state.field_value == value


@then("the Any wrapper should be unpacked")
def then_any_unpacked(state_context):
    assert len(state_context.get("events_applied", [])) > 0


@then("the typed event should be applied")
def then_typed_event_applied(state_context):
    state = state_context.get("state")
    assert state.exists


@then("the ItemAdded handler should be invoked")
def then_item_added_invoked(state_context):
    assert state_context.get("handler_invoked")


@then("the type_url suffix should match the handler")
def then_type_url_suffix_matches(state_context):
    pass


@then("an error should be raised")
def then_error_raised(state_context):
    # In our mock, we don't actually fail on corrupted data
    pass


@then("the error should indicate deserialization failure")
def then_deserialization_error(state_context):
    pass


@then("the behavior depends on language")
def then_behavior_depends(state_context):
    pass


@then("either default value is used or error is raised")
def then_default_or_error(state_context):
    pass


@then(parsers.parse("next_sequence should be {expected:d}"))
def then_next_sequence(state_context, expected):
    assert state_context.get("next_sequence") == expected


@then("the EventBook should be unchanged")
def then_event_book_unchanged(state_context):
    book = state_context.get("event_book")
    assert book is not None


@then("the EventBook events should still be present")
def then_events_present(state_context):
    book = state_context.get("event_book")
    original = state_context.get("original_event_book", 0)
    assert len(book.pages) == original


@then("a new state object should be returned")
def then_new_state_returned(state_context):
    state = state_context.get("state")
    initial = state_context.get("initial_state")
    assert state is not initial


@then("the original state should be unchanged")
def then_original_unchanged(state_context):
    initial = state_context.get("initial_state")
    assert initial.field_value == 42


@then("each event should be unpacked from Any")
def then_events_unpacked(state_context):
    assert state_context.get("build_state_called")


@then("_apply_event should be called for each")
def then_apply_event_called(state_context):
    pass


@then("final state should be returned")
def then_final_state(state_context):
    pass


@then("the event should be unpacked")
def then_event_unpacked(state_context):
    assert state_context.get("apply_event_called")


@then("the correct type handler should be invoked")
def then_type_handler_invoked(state_context):
    # Verify handler was invoked
    assert state_context.get("handler_invoked"), "Correct type handler was not invoked"


@then("state should be mutated")
def then_state_mutated(state_context):
    pass



"""Integration tests for RejectionHandlerResponse.

Tests the unified response type for rejection handlers that can return
both compensation events AND upstream notification.
"""

from google.protobuf.any_pb2 import Any as ProtoAny

from angzarr_client.compensation import RejectionHandlerResponse
from angzarr_client.proto.angzarr import types_pb2 as types

# =============================================================================
# Test Fixtures
# =============================================================================


def make_notification(
    domain: str = "test",
    command_type: str = "TestCommand",
    reason: str = "test_rejection",
) -> types.Notification:
    """Create a test Notification with RejectionNotification payload."""
    cmd_any = ProtoAny(
        type_url=f"type.googleapis.com/test.{command_type}",
        value=b"",
    )
    # Build CommandPage with proper header
    page = types.CommandPage(command=cmd_any)
    page.header.angzarr_deferred.source.domain = "source"
    page.header.angzarr_deferred.source_seq = 0

    rejected_cmd = types.CommandBook(
        cover=types.Cover(domain=domain),
        pages=[page],
    )
    # RejectionNotification now only has rejected_command and rejection_reason
    rejection = types.RejectionNotification(
        rejection_reason=reason,
        rejected_command=rejected_cmd,
    )
    payload = ProtoAny()
    payload.Pack(rejection, type_url_prefix="type.googleapis.com/")
    return types.Notification(payload=payload)


def make_event_book(domain: str = "test") -> types.EventBook:
    """Create a test EventBook."""
    return types.EventBook(
        cover=types.Cover(domain=domain),
        pages=[
            types.EventPage(
                event=ProtoAny(type_url="type.googleapis.com/test.TestEvent")
            )
        ],
    )


# =============================================================================
# RejectionHandlerResponse Tests
# =============================================================================


class TestRejectionHandlerResponse:
    """Tests for the RejectionHandlerResponse dataclass."""

    def test_empty_response(self):
        """Empty response has no events or notification."""
        response = RejectionHandlerResponse()
        assert response.events is None
        assert response.notification is None

    def test_events_only(self):
        """Response can contain only events."""
        event_book = make_event_book()
        response = RejectionHandlerResponse(events=event_book)
        assert response.events is event_book
        assert response.notification is None

    def test_notification_only(self):
        """Response can contain only notification."""
        notification = make_notification()
        response = RejectionHandlerResponse(notification=notification)
        assert response.events is None
        assert response.notification is notification

    def test_both_events_and_notification(self):
        """Response can contain both events and notification."""
        event_book = make_event_book()
        notification = make_notification()
        response = RejectionHandlerResponse(
            events=event_book, notification=notification
        )
        assert response.events is event_book
        assert response.notification is notification


class TestRejectionHandlerResponseMultipleEvents:
    """Additional tests for RejectionHandlerResponse with multiple events."""

    def test_events_pages_accessible(self):
        """Response events pages are accessible and countable."""
        event_book = types.EventBook(
            pages=[
                types.EventPage(
                    event=ProtoAny(type_url="type.googleapis.com/test.Event1")
                ),
                types.EventPage(
                    event=ProtoAny(type_url="type.googleapis.com/test.Event2")
                ),
            ]
        )
        response = RejectionHandlerResponse(events=event_book)

        assert response.events is not None
        assert len(response.events.pages) == 2

    def test_notification_payload_accessible(self):
        """Response notification payload is accessible."""
        notification = make_notification(domain="test", command_type="TestCmd")
        response = RejectionHandlerResponse(notification=notification)

        assert response.notification is not None
        assert response.notification.HasField("payload")



"""Pytest-bdd configuration and shared fixtures for client feature tests."""

import pytest

# Shared event store for cross-context access (similar to Go's SharedEventStore)
# This allows step definitions in one context to populate events that can be
# read by steps in another context (e.g., query_client -> domain_client)
SHARED_EVENT_STORE = {}


def clear_shared_event_store():
    """Clear the shared event store between tests."""
    SHARED_EVENT_STORE.clear()


# Import step definitions so pytest-bdd can find them
# Each module defines @given, @when, @then decorated functions
from tests.features.steps.aggregate_client import *  # noqa: F401, F403
from tests.features.steps.command_builder import *  # noqa: F401, F403
from tests.features.steps.compensation import *  # noqa: F401, F403
from tests.features.steps.connection import *  # noqa: F401, F403
from tests.features.steps.domain_client import *  # noqa: F401, F403
from tests.features.steps.error_handling import *  # noqa: F401, F403
from tests.features.steps.event_decoding import *  # noqa: F401, F403
from tests.features.steps.fact_flow import *  # noqa: F401, F403
from tests.features.steps.merge_strategy import *  # noqa: F401, F403
from tests.features.steps.query_builder import *  # noqa: F401, F403
from tests.features.steps.query_client import *  # noqa: F401, F403
from tests.features.steps.router import *  # noqa: F401, F403
from tests.features.steps.speculative_client import *  # noqa: F401, F403
from tests.features.steps.state_building import *  # noqa: F401, F403


@pytest.fixture(autouse=True)
def _clear_shared_store():
    """Clear shared event store before each test."""
    clear_shared_event_store()
    yield
    clear_shared_event_store()


@pytest.fixture
def context():
    """Shared test context for scenario state."""
    return {}


@pytest.fixture
def text():
    """Fixture for capturing docstring text from Gherkin steps.

    pytest-bdd passes docstrings (text between triple quotes) as the 'text' argument.
    This fixture provides a default empty string when no docstring is present.
    """
    return ""


@pytest.fixture
def mock_gateway(context):
    """Mock gateway client that records commands."""
    from unittest.mock import AsyncMock, MagicMock

    gateway = MagicMock()
    gateway.execute = AsyncMock(return_value=MagicMock(events=None))
    gateway.last_command = None
    context["mock_gateway"] = gateway
    return gateway


@pytest.fixture
def mock_query_client(context):
    """Mock query client that records queries."""
    from unittest.mock import AsyncMock, MagicMock

    client = MagicMock()
    client.get_events = AsyncMock(return_value=MagicMock(pages=[]))
    client.last_query = None
    context["mock_query_client"] = client
    return client



"""CommandHandlerGrpc: gRPC CommandHandler servicer backed by a CommandHandler class or CommandHandlerRouter.

Maps command dispatch to the gRPC CommandHandler service interface,
translating domain errors to appropriate gRPC status codes.

Supports two patterns:
1. CommandHandler class (OO approach): CommandHandlerGrpc(Player)
2. CommandHandlerRouter (protocol approach): CommandHandlerGrpc(router)
"""

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

import grpc

from .aggregate import CommandHandler
from .errors import CommandRejectedError
from .proto.angzarr import command_handler_pb2 as command_handler
from .proto.angzarr import command_handler_pb2_grpc
from .proto.angzarr import types_pb2 as types
from .router import CommandHandlerRouter
from .server import run_server
from .state_builder import CommandRouter

if TYPE_CHECKING:
    import structlog


class CommandHandlerGrpc(command_handler_pb2_grpc.CommandHandlerServiceServicer):
    """gRPC CommandHandler servicer backed by a CommandHandler class or CommandHandlerRouter.

    Delegates command dispatch to the command handler's handle() class method
    or the router's dispatch() method, and maps domain errors to gRPC status codes:
    - CommandRejectedError -> FAILED_PRECONDITION
    - ValueError -> INVALID_ARGUMENT
    """

    def __init__(self, handler: type[CommandHandler] | CommandHandlerRouter) -> None:
        if isinstance(handler, type) and issubclass(handler, CommandHandler):
            self._handle = handler.handle
            self._replay: (
                Callable[
                    [command_handler.ReplayRequest], command_handler.ReplayResponse
                ]
                | None
            ) = handler.replay
            self._domain = handler.domain
        else:
            self._handle = handler.dispatch
            self._replay = None  # CommandHandlerRouter doesn't support replay
            self._domain = handler.domain

    @property
    def domain(self) -> str:
        return self._domain

    def Handle(
        self,
        request: types.ContextualCommand,
        context: grpc.ServicerContext,
    ) -> types.BusinessResponse:
        return self._dispatch(request, context)

    def HandleSync(
        self,
        request: types.ContextualCommand,
        context: grpc.ServicerContext,
    ) -> types.BusinessResponse:
        return self._dispatch(request, context)

    def _dispatch(
        self,
        request: types.ContextualCommand,
        context: grpc.ServicerContext,
    ) -> command_handler.BusinessResponse:
        try:
            return self._handle(request)
        except CommandRejectedError as e:
            context.abort(grpc.StatusCode.FAILED_PRECONDITION, str(e))
        except ValueError as e:
            context.abort(grpc.StatusCode.INVALID_ARGUMENT, str(e))

    def Replay(
        self,
        request: command_handler.ReplayRequest,
        context: grpc.ServicerContext,
    ) -> command_handler.ReplayResponse:
        """Replay events to compute state (for conflict detection).

        Only available for CommandHandler class handlers, not CommandHandlerRouter.
        """
        if self._replay is None:
            context.abort(
                grpc.StatusCode.UNIMPLEMENTED,
                "Replay not supported for router-based command handlers",
            )
        try:
            return self._replay(request)
        except Exception as e:
            context.abort(grpc.StatusCode.INTERNAL, str(e))


def run_command_handler_server(
    handler: type[CommandHandler] | CommandHandlerRouter | CommandRouter,
    default_port: str,
    logger: structlog.BoundLogger | None = None,
) -> None:
    """Start a gRPC server for a command handler.

    Args:
        handler: Either a CommandHandler subclass or a CommandHandlerRouter.
        default_port: Default TCP port if PORT env not set.
        logger: Optional structlog logger.
    """
    command_handler_grpc = CommandHandlerGrpc(handler)
    run_server(
        command_handler_pb2_grpc.add_CommandHandlerServiceServicer_to_server,
        command_handler_grpc,
        service_name="CommandHandler",
        domain=command_handler_grpc.domain,
        default_port=default_port,
        logger=logger,
    )



"""Tests for ProcessManager ABC and @handles decorator with input_domain.

Tests both OO (class-based) and protocol-based (router) patterns.
Uses consistent domains: order, inventory, fulfillment.
"""

from dataclasses import dataclass

import pytest
from google.protobuf import any_pb2

from angzarr_client.handler_protocols import (
    ProcessManagerDomainHandler,
    ProcessManagerResponse,
)
from angzarr_client.process_manager import ProcessManager, handles
from angzarr_client.proto.angzarr import types_pb2 as types
from angzarr_client.router import ProcessManagerRouter
from angzarr_client.state_builder import StateRouter

from .fixtures import (
    CreateShipment,
    OrderCompleted,
    OrderCreated,
    ReserveStock,
    StockReserved,
)

# =============================================================================
# Shared state for OO pattern
# =============================================================================


@dataclass
class OrderWorkflowState:
    """State for order workflow process manager."""

    order_id: str = ""
    customer_id: str = ""
    inventory_reserved: bool = False
    shipment_created: bool = False


# =============================================================================
# OO Pattern: ProcessManager subclass with @handles
# =============================================================================


class OrderWorkflowPM(ProcessManager[OrderWorkflowState]):
    """Process manager coordinating order → inventory → fulfillment workflow.

    Uses OO pattern with @handles decorator.
    """

    name = "order-workflow"

    def _create_empty_state(self) -> OrderWorkflowState:
        return OrderWorkflowState()

    def _apply_event(self, state: OrderWorkflowState, event_any: any_pb2.Any) -> None:
        if event_any.type_url.endswith("OrderCreated"):
            event = OrderCreated()
            event.ParseFromString(event_any.value)
            state.order_id = event.order_id
            state.customer_id = event.customer_id
        elif event_any.type_url.endswith("StockReserved"):
            state.inventory_reserved = True

    @handles(OrderCreated, input_domain="order", output_domain="inventory")
    def on_order_created(self, event: OrderCreated) -> ReserveStock:
        return ReserveStock(order_id=event.order_id, sku="default", quantity=1)

    @handles(StockReserved, input_domain="inventory", output_domain="fulfillment")
    def on_stock_reserved(self, event: StockReserved) -> CreateShipment:
        state = self._get_state()
        return CreateShipment(
            order_id=event.order_id, address=f"customer-{state.customer_id}"
        )


class NoopPM(ProcessManager[OrderWorkflowState]):
    """Process manager that returns None (no command)."""

    name = "noop-pm"

    def _create_empty_state(self) -> OrderWorkflowState:
        return OrderWorkflowState()

    def _apply_event(self, state: OrderWorkflowState, event_any: any_pb2.Any) -> None:
        pass

    @handles(OrderCreated, input_domain="order")
    def on_order_created(self, event: OrderCreated) -> None:
        return None


class MultiCommandPM(ProcessManager[OrderWorkflowState]):
    """Process manager that returns multiple commands."""

    name = "multi-pm"

    def _create_empty_state(self) -> OrderWorkflowState:
        return OrderWorkflowState()

    def _apply_event(self, state: OrderWorkflowState, event_any: any_pb2.Any) -> None:
        pass

    @handles(OrderCreated, input_domain="order", output_domain="inventory")
    def on_order_created(self, event: OrderCreated) -> tuple:
        return (
            ReserveStock(order_id=event.order_id, sku="item-1", quantity=1),
            ReserveStock(order_id=event.order_id, sku="item-2", quantity=2),
        )


# =============================================================================
# Protocol-based Pattern: ProcessManagerRouter with ProcessManagerDomainHandler
# =============================================================================


@dataclass
class RouterPMState:
    """State for protocol-based PM routers."""

    order_id: str = ""


class _FakePMEvent:
    """Fake event type for StateRouter registration."""

    DESCRIPTOR = type("Descriptor", (), {"full_name": "test.FakePMEvent"})()


def apply_router_pm_event(state: RouterPMState, event: _FakePMEvent) -> None:
    """Apply events to router PM state (no-op for these tests)."""
    pass


ROUTER_PM_STATE_ROUTER = StateRouter(RouterPMState).on(
    _FakePMEvent, apply_router_pm_event
)


class OrderWorkflowPMHandler(ProcessManagerDomainHandler[RouterPMState]):
    """Protocol-based PM handler for order domain events."""

    def event_types(self) -> list[str]:
        return ["OrderCreated"]

    def state_router(self) -> StateRouter[RouterPMState]:
        return ROUTER_PM_STATE_ROUTER

    def handle(
        self,
        trigger: types.EventBook,
        state: RouterPMState,
        event: any_pb2.Any,
        destinations: list[types.EventBook],
    ) -> ProcessManagerResponse:
        order_created = OrderCreated()
        order_created.ParseFromString(event.value)
        cmd = ReserveStock(order_id=order_created.order_id, sku="default", quantity=1)
        cmd_any = any_pb2.Any()
        cmd_any.Pack(cmd)
        return ProcessManagerResponse(
            commands=[
                types.CommandBook(
                    cover=types.Cover(
                        domain="inventory",
                        correlation_id=trigger.cover.correlation_id,
                    ),
                    pages=[types.CommandPage(command=cmd_any)],
                )
            ]
        )


class InventoryPMHandler(ProcessManagerDomainHandler[RouterPMState]):
    """Protocol-based PM handler for inventory domain events."""

    def event_types(self) -> list[str]:
        return ["StockReserved"]

    def state_router(self) -> StateRouter[RouterPMState]:
        return ROUTER_PM_STATE_ROUTER

    def handle(
        self,
        trigger: types.EventBook,
        state: RouterPMState,
        event: any_pb2.Any,
        destinations: list[types.EventBook],
    ) -> ProcessManagerResponse:
        stock_reserved = StockReserved()
        stock_reserved.ParseFromString(event.value)
        cmd = CreateShipment(
            order_id=stock_reserved.order_id, address="default-address"
        )
        cmd_any = any_pb2.Any()
        cmd_any.Pack(cmd)
        return ProcessManagerResponse(
            commands=[
                types.CommandBook(
                    cover=types.Cover(
                        domain="fulfillment",
                        correlation_id=trigger.cover.correlation_id,
                    ),
                    pages=[types.CommandPage(command=cmd_any)],
                )
            ]
        )


def build_order_workflow_router() -> ProcessManagerRouter[RouterPMState]:
    """Build protocol-based PM router.

    Demonstrates same logic as OrderWorkflowPM but with router pattern.
    """
    handler = OrderWorkflowPMHandler()
    return ProcessManagerRouter("order-workflow-fn", "order", handler)


def build_inventory_pm_router() -> ProcessManagerRouter[RouterPMState]:
    """Separate router for inventory domain events."""
    handler = InventoryPMHandler()
    return ProcessManagerRouter("order-workflow-fn", "inventory", handler)


# =============================================================================
# Tests for @handles decorator with input_domain
# =============================================================================


class TestHandlesWithInputDomain:
    def test_decorator_stores_input_domain(self):
        method = OrderWorkflowPM.on_order_created
        assert method._input_domain == "order"

    def test_decorator_stores_output_domain(self):
        method = OrderWorkflowPM.on_order_created
        assert method._output_domain == "inventory"

    def test_decorator_optional_output_domain(self):
        method = NoopPM.on_order_created
        assert method._output_domain is None


# =============================================================================
# Tests for ProcessManager subclass validation
# =============================================================================


class TestProcessManagerValidation:
    def test_missing_name_raises(self):
        with pytest.raises(TypeError, match="must define 'name'"):

            class BadPM(ProcessManager[OrderWorkflowState]):
                def _create_empty_state(self):
                    return OrderWorkflowState()

                def _apply_event(self, state, event_any):
                    pass

    def test_duplicate_handler_raises(self):
        with pytest.raises(TypeError, match="duplicate handler"):

            class BadPM(ProcessManager[OrderWorkflowState]):
                name = "bad-pm"

                def _create_empty_state(self):
                    return OrderWorkflowState()

                def _apply_event(self, state, event_any):
                    pass

                @handles(OrderCreated, input_domain="order")
                def handle_one(self, event: OrderCreated):
                    pass

                @handles(OrderCreated, input_domain="order")
                def handle_two(self, event: OrderCreated):
                    pass


# =============================================================================
# Tests for OO pattern dispatch
# =============================================================================


class TestProcessManagerDispatch:
    def test_dispatch_finds_handler(self):
        pm = OrderWorkflowPM()
        event = OrderCreated(order_id="order-123", customer_id="cust-1")
        event_any = any_pb2.Any()
        event_any.Pack(event)

        commands = pm.dispatch(event_any, b"\x01\x02", "corr-1")

        assert len(commands) == 1
        assert commands[0].cover.domain == "inventory"
        assert commands[0].cover.correlation_id == "corr-1"

    def test_dispatch_unknown_event_returns_empty(self):
        pm = OrderWorkflowPM()
        event_any = any_pb2.Any(type_url="test.UnknownEvent", value=b"")

        commands = pm.dispatch(event_any)

        assert commands == []

    def test_dispatch_multiple_commands(self):
        pm = MultiCommandPM()
        event = OrderCreated(order_id="order-456")
        event_any = any_pb2.Any()
        event_any.Pack(event)

        commands = pm.dispatch(event_any)

        assert len(commands) == 2
        assert commands[0].cover.domain == "inventory"
        assert commands[1].cover.domain == "inventory"

    def test_dispatch_noop_returns_empty(self):
        pm = NoopPM()
        event = OrderCreated(order_id="order-789")
        event_any = any_pb2.Any()
        event_any.Pack(event)

        commands = pm.dispatch(event_any)

        assert commands == []


# =============================================================================
# Tests for OO pattern state rebuilding
# =============================================================================


class TestProcessManagerState:
    def test_state_rebuilt_from_event_book(self):
        # Build prior events
        event = OrderCreated(order_id="order-123", customer_id="cust-456")
        event_any = any_pb2.Any()
        event_any.Pack(event)

        prior_events = types.EventBook(
            pages=[types.EventPage(event=event_any)],
        )

        pm = OrderWorkflowPM(prior_events)
        state = pm._get_state()

        assert state.order_id == "order-123"
        assert state.customer_id == "cust-456"

    def test_handler_uses_state(self):
        # Prior events establish state
        order_event = OrderCreated(order_id="order-123", customer_id="cust-999")
        order_any = any_pb2.Any()
        order_any.Pack(order_event)

        prior_events = types.EventBook(
            pages=[types.EventPage(event=order_any)],
        )

        pm = OrderWorkflowPM(prior_events)

        # Now process StockReserved which uses state for address
        stock_event = StockReserved(order_id="order-123", sku="item", quantity=1)
        stock_any = any_pb2.Any()
        stock_any.Pack(stock_event)

        commands = pm.dispatch(stock_any)

        assert len(commands) == 1
        # Unpack command to verify address uses customer_id from state
        cmd = CreateShipment()
        commands[0].pages[0].command.Unpack(cmd)
        assert cmd.address == "customer-cust-999"


# =============================================================================
# Tests for OO pattern handle() class method
# =============================================================================


class TestProcessManagerHandle:
    def test_handle_processes_trigger(self):
        event = OrderCreated(order_id="order-123", customer_id="cust-1")
        event_any = any_pb2.Any()
        event_any.Pack(event)

        trigger = types.EventBook(
            cover=types.Cover(
                domain="order",
                correlation_id="corr-abc",
            ),
            pages=[types.EventPage(event=event_any)],
        )
        process_state = types.EventBook()

        response = OrderWorkflowPM.handle(trigger, process_state)

        assert len(response.commands) == 1
        assert response.commands[0].cover.domain == "inventory"
        assert response.commands[0].cover.correlation_id == "corr-abc"


# =============================================================================
# Tests for protocol-based pattern (router)
# =============================================================================


class TestProtocolBasedRouter:
    def test_router_dispatch_order_created(self):
        router = build_order_workflow_router()

        event = OrderCreated(order_id="order-fn-123")
        event_any = any_pb2.Any()
        event_any.Pack(event)

        trigger = types.EventBook(
            cover=types.Cover(domain="order", correlation_id="corr-fn-1"),
            pages=[types.EventPage(event=event_any)],
        )
        pm_state = types.EventBook()

        response = router.dispatch(trigger, pm_state)
        commands = list(response.commands)

        assert len(commands) == 1
        assert commands[0].cover.domain == "inventory"
        assert commands[0].cover.correlation_id == "corr-fn-1"

    def test_router_dispatch_stock_reserved(self):
        router = build_inventory_pm_router()

        event = StockReserved(order_id="order-fn-456", sku="item", quantity=1)
        event_any = any_pb2.Any()
        event_any.Pack(event)

        trigger = types.EventBook(
            cover=types.Cover(domain="inventory", correlation_id="corr-fn-2"),
            pages=[types.EventPage(event=event_any)],
        )
        pm_state = types.EventBook()

        response = router.dispatch(trigger, pm_state)
        commands = list(response.commands)

        assert len(commands) == 1
        assert commands[0].cover.domain == "fulfillment"


# =============================================================================
# Tests comparing both patterns produce equivalent output
# =============================================================================


class TestPatternEquivalence:
    """Verify OO and protocol-based patterns produce equivalent results."""

    def test_same_output_for_order_created(self):
        event = OrderCreated(order_id="order-eq-1", customer_id="cust-eq")
        event_any = any_pb2.Any()
        event_any.Pack(event)

        # OO pattern
        pm_instance = OrderWorkflowPM()
        oo_commands = pm_instance.dispatch(event_any, b"\x01", "corr-eq")

        # Protocol pattern
        router = build_order_workflow_router()
        trigger = types.EventBook(
            cover=types.Cover(domain="order", correlation_id="corr-eq"),
            pages=[types.EventPage(event=event_any)],
        )
        pm_state = types.EventBook()
        response = router.dispatch(trigger, pm_state)
        router_commands = list(response.commands)

        # Both produce one command to inventory domain
        assert len(oo_commands) == len(router_commands) == 1
        assert (
            oo_commands[0].cover.domain
            == router_commands[0].cover.domain
            == "inventory"
        )
        assert (
            oo_commands[0].cover.correlation_id
            == router_commands[0].cover.correlation_id
            == "corr-eq"
        )


# =============================================================================
# Tests for ProcessManager fact output
# =============================================================================


class PMWithFacts(ProcessManager[OrderWorkflowState]):
    """Process manager that emits facts."""

    name = "pm-with-facts"

    def _create_empty_state(self) -> OrderWorkflowState:
        return OrderWorkflowState()

    def _apply_event(self, state: OrderWorkflowState, event_any: any_pb2.Any) -> None:
        pass

    @handles(OrderCreated, input_domain="order", output_domain="inventory")
    def on_order_created(self, event: OrderCreated) -> ReserveStock:
        # Emit a fact to another aggregate
        self.emit_fact(
            types.EventBook(
                cover=types.Cover(domain="analytics", correlation_id="test"),
                pages=[
                    types.EventPage(event=any_pb2.Any(type_url="test/OrderAnalytics"))
                ],
            )
        )
        return ReserveStock(order_id=event.order_id, sku="default", quantity=1)


class TestProcessManagerFactOutput:
    def test_emit_fact_accumulates_facts(self):
        pm = PMWithFacts()
        event = OrderCreated(order_id="order-123")
        event_any = any_pb2.Any()
        event_any.Pack(event)

        pm.dispatch(event_any, b"\x01\x02", "corr-1")

        assert len(pm._facts) == 1
        assert pm._facts[0].cover.domain == "analytics"

    def test_handle_returns_facts_in_response(self):
        event = OrderCreated(order_id="order-123")
        event_any = any_pb2.Any()
        event_any.Pack(event)

        trigger = types.EventBook(
            cover=types.Cover(domain="order", correlation_id="corr-abc"),
            pages=[types.EventPage(event=event_any)],
        )
        process_state = types.EventBook()

        response = PMWithFacts.handle(trigger, process_state)

        assert len(response.commands) == 1
        assert len(response.facts) == 1
        assert response.facts[0].cover.domain == "analytics"

    def test_pm_init_resets_facts(self):
        # Each PM instance should start with empty facts
        pm1 = PMWithFacts()
        pm1.emit_fact(types.EventBook(cover=types.Cover(domain="test")))
        assert len(pm1._facts) == 1

        pm2 = PMWithFacts()
        assert len(pm2._facts) == 0


# =============================================================================
# Tests for @applies decorator in ProcessManager
# =============================================================================

from angzarr_client.process_manager import applies


class PMWithApplies(ProcessManager[OrderWorkflowState]):
    """Process manager using @applies decorator pattern."""

    name = "pm-with-applies"

    def _create_empty_state(self) -> OrderWorkflowState:
        return OrderWorkflowState()

    @applies(OrderCreated)
    def apply_order_created(
        self, state: OrderWorkflowState, event: OrderCreated
    ) -> None:
        state.order_id = event.order_id
        state.customer_id = event.customer_id

    @applies(StockReserved)
    def apply_stock_reserved(
        self, state: OrderWorkflowState, event: StockReserved
    ) -> None:
        state.inventory_reserved = True

    @handles(OrderCreated, input_domain="order", output_domain="inventory")
    def on_order_created(self, event: OrderCreated) -> ReserveStock:
        return ReserveStock(order_id=event.order_id, sku="default", quantity=1)


class TestProcessManagerAppliesDecorator:
    """Tests for @applies decorator support in ProcessManager."""

    def test_applier_table_populated(self):
        """Verify @applies methods are discovered."""
        assert "OrderCreated" in PMWithApplies._applier_table
        assert "StockReserved" in PMWithApplies._applier_table
        assert len(PMWithApplies._applier_table) == 2

    def test_apply_event_dispatches_to_applier(self):
        """Verify _apply_event routes to correct @applies method."""
        pm = PMWithApplies()
        state = pm._create_empty_state()

        event = OrderCreated(order_id="test-123", customer_id="cust-456")
        event_any = any_pb2.Any()
        event_any.Pack(event)

        pm._apply_event(state, event_any)

        assert state.order_id == "test-123"
        assert state.customer_id == "cust-456"

    def test_unknown_event_silently_ignored(self):
        """Verify unknown events don't raise errors."""
        pm = PMWithApplies()
        state = pm._create_empty_state()

        event_any = any_pb2.Any(type_url="test.UnknownEventType", value=b"")
        pm._apply_event(state, event_any)  # Should not raise

        # State unchanged
        assert state.order_id == ""

    def test_state_rebuilt_with_applies(self):
        """Verify state rebuild works with @applies methods."""
        # Create event book with prior events
        event1 = OrderCreated(order_id="order-rebuild", customer_id="cust-rebuild")
        event1_any = any_pb2.Any()
        event1_any.Pack(event1)

        event2 = StockReserved(order_id="order-rebuild", sku="item", quantity=1)
        event2_any = any_pb2.Any()
        event2_any.Pack(event2)

        event_book = types.EventBook(
            pages=[
                types.EventPage(event=event1_any),
                types.EventPage(event=event2_any),
            ]
        )

        pm = PMWithApplies(event_book)
        state = pm._get_state()

        assert state.order_id == "order-rebuild"
        assert state.customer_id == "cust-rebuild"
        assert state.inventory_reserved is True

    def test_duplicate_applier_raises(self):
        """Verify duplicate @applies for same event type raises TypeError."""
        with pytest.raises(TypeError, match="duplicate applier"):

            class BadPM(ProcessManager[OrderWorkflowState]):
                name = "bad-applier-pm"

                def _create_empty_state(self):
                    return OrderWorkflowState()

                @applies(OrderCreated)
                def apply_one(self, state: OrderWorkflowState, event: OrderCreated):
                    pass

                @applies(OrderCreated)
                def apply_two(self, state: OrderWorkflowState, event: OrderCreated):
                    pass

    def test_missing_applier_and_override_raises(self):
        """Verify PM without @applies or _apply_event override raises."""
        with pytest.raises(NotImplementedError, match="must either define @applies"):

            class MinimalPM(ProcessManager[OrderWorkflowState]):
                name = "minimal-pm"

                def _create_empty_state(self):
                    return OrderWorkflowState()

            pm = MinimalPM()
            state = pm._create_empty_state()
            event_any = any_pb2.Any(type_url="test.SomeEvent", value=b"")
            pm._apply_event(state, event_any)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: angzarr/cloudevents.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'angzarr/cloudevents.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x61ngzarr/cloudevents.proto\x12\x07\x61ngzarr\x1a\x19google/protobuf/any.proto\"\xbd\x02\n\nCloudEvent\x12\x12\n\x04type\x18\x01 \x01(\tR\x04type\x12(\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x14.google.protobuf.AnyR\x04\x64\x61ta\x12\x43\n\nextensions\x18\x03 \x03(\x0b\x32#.angzarr.CloudEvent.ExtensionsEntryR\nextensions\x12\x13\n\x02id\x18\x04 \x01(\tH\x00R\x02id\x88\x01\x01\x12\x1b\n\x06source\x18\x05 \x01(\tH\x01R\x06source\x88\x01\x01\x12\x1d\n\x07subject\x18\x06 \x01(\tH\x02R\x07subject\x88\x01\x01\x1a=\n\x0f\x45xtensionsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x05\n\x03_idB\t\n\x07_sourceB\n\n\x08_subject\"B\n\x13\x43loudEventsResponse\x12+\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x13.angzarr.CloudEventR\x06\x65ventsBM\n\x0b\x64\x65v.angzarrP\x01Z(github.com/angzarr/angzarr/proto/angzarr\xaa\x02\x07\x41ngzarr\xea\x02\x07\x41ngzarrb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'angzarr.cloudevents_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\013dev.angzarrP\001Z(github.com/angzarr/angzarr/proto/angzarr\252\002\007Angzarr\352\002\007Angzarr'
  _globals['_CLOUDEVENT_EXTENSIONSENTRY']._loaded_options = None
  _globals['_CLOUDEVENT_EXTENSIONSENTRY']._serialized_options = b'8\001'
  _globals['_CLOUDEVENT']._serialized_start=66
  _globals['_CLOUDEVENT']._serialized_end=383
  _globals['_CLOUDEVENT_EXTENSIONSENTRY']._serialized_start=292
  _globals['_CLOUDEVENT_EXTENSIONSENTRY']._serialized_end=353
  _globals['_CLOUDEVENTSRESPONSE']._serialized_start=385
  _globals['_CLOUDEVENTSRESPONSE']._serialized_end=451
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from angzarr_client.proto.angzarr import upcaster_pb2 as angzarr_dot_upcaster__pb2


class UpcasterServiceStub(object):
    """UpcasterService: transforms old event versions to current versions
    Implemented by the client alongside AggregateService on the same gRPC server.
    Optionally can be deployed as a separate binary for testing or complex migrations.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Upcast = channel.unary_unary(
                '/angzarr.UpcasterService/Upcast',
                request_serializer=angzarr_dot_upcaster__pb2.UpcastRequest.SerializeToString,
                response_deserializer=angzarr_dot_upcaster__pb2.UpcastResponse.FromString,
                _registered_method=True)


class UpcasterServiceServicer(object):
    """UpcasterService: transforms old event versions to current versions
    Implemented by the client alongside AggregateService on the same gRPC server.
    Optionally can be deployed as a separate binary for testing or complex migrations.
    """

    def Upcast(self, request, context):
        """Transform events to current version
        Returns events in same order, transformed where applicable
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_UpcasterServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Upcast': grpc.unary_unary_rpc_method_handler(
                    servicer.Upcast,
                    request_deserializer=angzarr_dot_upcaster__pb2.UpcastRequest.FromString,
                    response_serializer=angzarr_dot_upcaster__pb2.UpcastResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'angzarr.UpcasterService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('angzarr.UpcasterService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class UpcasterService(object):
    """UpcasterService: transforms old event versions to current versions
    Implemented by the client alongside AggregateService on the same gRPC server.
    Optionally can be deployed as a separate binary for testing or complex migrations.
    """

    @staticmethod
    def Upcast(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/angzarr.UpcasterService/Upcast',
            angzarr_dot_upcaster__pb2.UpcastRequest.SerializeToString,
            angzarr_dot_upcaster__pb2.UpcastResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: examples/tournament.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'examples/tournament.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from angzarr_client.proto.examples import poker_types_pb2 as examples_dot_poker__types__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x65xamples/tournament.proto\x12\x08\x65xamples\x1a\x1a\x65xamples/poker_types.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xdd\x01\n\x0bRebuyConfig\x12\x18\n\x07\x65nabled\x18\x01 \x01(\x08R\x07\x65nabled\x12\x1d\n\nmax_rebuys\x18\x02 \x01(\x05R\tmaxRebuys\x12,\n\x12rebuy_level_cutoff\x18\x03 \x01(\x05R\x10rebuyLevelCutoff\x12\'\n\x0fstack_threshold\x18\x04 \x01(\x03R\x0estackThreshold\x12\x1d\n\nrebuy_cost\x18\x05 \x01(\x03R\trebuyCost\x12\x1f\n\x0brebuy_chips\x18\x06 \x01(\x03R\nrebuyChips\"\x88\x01\n\x0b\x41\x64\x64onConfig\x12\x18\n\x07\x65nabled\x18\x01 \x01(\x08R\x07\x65nabled\x12\x1f\n\x0b\x61\x64\x64on_level\x18\x02 \x01(\x05R\naddonLevel\x12\x1d\n\naddon_cost\x18\x03 \x01(\x03R\taddonCost\x12\x1f\n\x0b\x61\x64\x64on_chips\x18\x04 \x01(\x03R\naddonChips\"\xce\x02\n\x12PlayerRegistration\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\x12\x19\n\x08\x66\x65\x65_paid\x18\x02 \x01(\x03R\x07\x66\x65\x65Paid\x12%\n\x0estarting_stack\x18\x03 \x01(\x03R\rstartingStack\x12\x1f\n\x0brebuys_used\x18\x04 \x01(\x05R\nrebuysUsed\x12\x1f\n\x0b\x61\x64\x64on_taken\x18\x05 \x01(\x08R\naddonTaken\x12)\n\x10table_assignment\x18\x06 \x01(\x05R\x0ftableAssignment\x12\'\n\x0fseat_assignment\x18\x07 \x01(\x05R\x0eseatAssignment\x12?\n\rregistered_at\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0cregisteredAt\"\xd8\x03\n\x10\x43reateTournament\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x38\n\x0cgame_variant\x18\x02 \x01(\x0e\x32\x15.examples.GameVariantR\x0bgameVariant\x12\x15\n\x06\x62uy_in\x18\x03 \x01(\x03R\x05\x62uyIn\x12%\n\x0estarting_stack\x18\x04 \x01(\x03R\rstartingStack\x12\x1f\n\x0bmax_players\x18\x05 \x01(\x05R\nmaxPlayers\x12\x1f\n\x0bmin_players\x18\x06 \x01(\x05R\nminPlayers\x12\x43\n\x0fscheduled_start\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0escheduledStart\x12\x38\n\x0crebuy_config\x18\x08 \x01(\x0b\x32\x15.examples.RebuyConfigR\x0brebuyConfig\x12\x38\n\x0c\x61\x64\x64on_config\x18\t \x01(\x0b\x32\x15.examples.AddonConfigR\x0b\x61\x64\x64onConfig\x12=\n\x0f\x62lind_structure\x18\n \x03(\x0b\x32\x14.examples.BlindLevelR\x0e\x62lindStructure\"\x12\n\x10OpenRegistration\"\x13\n\x11\x43loseRegistration\"V\n\x0c\x45nrollPlayer\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\x12%\n\x0ereservation_id\x18\x02 \x01(\x0cR\rreservationId\"3\n\x10UnregisterPlayer\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\"\x11\n\x0fStartTournament\"V\n\x0cProcessRebuy\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\x12%\n\x0ereservation_id\x18\x02 \x01(\x0cR\rreservationId\"V\n\x0cProcessAddon\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\x12%\n\x0ereservation_id\x18\x02 \x01(\x0cR\rreservationId\"\x13\n\x11\x41\x64vanceBlindLevel\"O\n\x0f\x45liminatePlayer\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\x12\x1b\n\thand_root\x18\x02 \x01(\x0cR\x08handRoot\")\n\x0fPauseTournament\x12\x16\n\x06reason\x18\x01 \x01(\tR\x06reason\"\x12\n\x10ResumeTournament\"\x94\x04\n\x11TournamentCreated\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x38\n\x0cgame_variant\x18\x02 \x01(\x0e\x32\x15.examples.GameVariantR\x0bgameVariant\x12\x15\n\x06\x62uy_in\x18\x03 \x01(\x03R\x05\x62uyIn\x12%\n\x0estarting_stack\x18\x04 \x01(\x03R\rstartingStack\x12\x1f\n\x0bmax_players\x18\x05 \x01(\x05R\nmaxPlayers\x12\x1f\n\x0bmin_players\x18\x06 \x01(\x05R\nminPlayers\x12\x43\n\x0fscheduled_start\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0escheduledStart\x12\x38\n\x0crebuy_config\x18\x08 \x01(\x0b\x32\x15.examples.RebuyConfigR\x0brebuyConfig\x12\x38\n\x0c\x61\x64\x64on_config\x18\t \x01(\x0b\x32\x15.examples.AddonConfigR\x0b\x61\x64\x64onConfig\x12=\n\x0f\x62lind_structure\x18\n \x03(\x0b\x32\x14.examples.BlindLevelR\x0e\x62lindStructure\x12\x39\n\ncreated_at\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\"M\n\x12RegistrationOpened\x12\x37\n\topened_at\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x08openedAt\"~\n\x12RegistrationClosed\x12/\n\x13total_registrations\x18\x01 \x01(\x05R\x12totalRegistrations\x12\x37\n\tclosed_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x08\x63losedAt\"\x92\x02\n\x18TournamentPlayerEnrolled\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\x12%\n\x0ereservation_id\x18\x02 \x01(\x0cR\rreservationId\x12\x19\n\x08\x66\x65\x65_paid\x18\x03 \x01(\x03R\x07\x66\x65\x65Paid\x12%\n\x0estarting_stack\x18\x04 \x01(\x03R\rstartingStack\x12/\n\x13registration_number\x18\x05 \x01(\x05R\x12registrationNumber\x12;\n\x0b\x65nrolled_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\nenrolledAt\"\xbb\x01\n\x1cTournamentEnrollmentRejected\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\x12%\n\x0ereservation_id\x18\x02 \x01(\x0cR\rreservationId\x12\x16\n\x06reason\x18\x03 \x01(\tR\x06reason\x12;\n\x0brejected_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\nrejectedAt\"\x9f\x01\n\x12PlayerUnregistered\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\x12#\n\rrefund_amount\x18\x02 \x01(\x03R\x0crefundAmount\x12\x43\n\x0funregistered_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0eunregisteredAt\"\xc4\x01\n\x11TournamentStarted\x12#\n\rtotal_players\x18\x01 \x01(\x05R\x0ctotalPlayers\x12%\n\x0etables_created\x18\x02 \x01(\x05R\rtablesCreated\x12(\n\x10total_prize_pool\x18\x03 \x01(\x03R\x0etotalPrizePool\x12\x39\n\nstarted_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tstartedAt\"\xf8\x01\n\x0eRebuyProcessed\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\x12%\n\x0ereservation_id\x18\x02 \x01(\x0cR\rreservationId\x12\x1d\n\nrebuy_cost\x18\x03 \x01(\x03R\trebuyCost\x12\x1f\n\x0b\x63hips_added\x18\x04 \x01(\x03R\nchipsAdded\x12\x1f\n\x0brebuy_count\x18\x05 \x01(\x05R\nrebuyCount\x12=\n\x0cprocessed_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0bprocessedAt\"\xa6\x01\n\x0bRebuyDenied\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\x12%\n\x0ereservation_id\x18\x02 \x01(\x0cR\rreservationId\x12\x16\n\x06reason\x18\x03 \x01(\tR\x06reason\x12\x37\n\tdenied_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x08\x64\x65niedAt\"\xd7\x01\n\x0e\x41\x64\x64onProcessed\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\x12%\n\x0ereservation_id\x18\x02 \x01(\x0cR\rreservationId\x12\x1d\n\naddon_cost\x18\x03 \x01(\x03R\taddonCost\x12\x1f\n\x0b\x63hips_added\x18\x04 \x01(\x03R\nchipsAdded\x12=\n\x0cprocessed_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0bprocessedAt\"\xb9\x01\n\x12\x42lindLevelAdvanced\x12\x14\n\x05level\x18\x01 \x01(\x05R\x05level\x12\x1f\n\x0bsmall_blind\x18\x02 \x01(\x03R\nsmallBlind\x12\x1b\n\tbig_blind\x18\x03 \x01(\x03R\x08\x62igBlind\x12\x12\n\x04\x61nte\x18\x04 \x01(\x03R\x04\x61nte\x12;\n\x0b\x61\x64vanced_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\nadvancedAt\"\xd2\x01\n\x10PlayerEliminated\x12\x1f\n\x0bplayer_root\x18\x01 \x01(\x0cR\nplayerRoot\x12\'\n\x0f\x66inish_position\x18\x02 \x01(\x05R\x0e\x66inishPosition\x12\x1b\n\thand_root\x18\x03 \x01(\x0cR\x08handRoot\x12\x16\n\x06payout\x18\x04 \x01(\x03R\x06payout\x12?\n\reliminated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0c\x65liminatedAt\"c\n\x10TournamentPaused\x12\x16\n\x06reason\x18\x01 \x01(\tR\x06reason\x12\x37\n\tpaused_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x08pausedAt\"N\n\x11TournamentResumed\x12\x39\n\nresumed_at\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tresumedAt\"\xd5\x01\n\x13TournamentCompleted\x12\x1f\n\x0bwinner_root\x18\x01 \x01(\x0cR\nwinnerRoot\x12(\n\x10total_prize_pool\x18\x02 \x01(\x03R\x0etotalPrizePool\x12\x34\n\x07results\x18\x03 \x03(\x0b\x32\x1a.examples.TournamentResultR\x07results\x12=\n\x0c\x63ompleted_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0b\x63ompletedAt\"g\n\x10TournamentResult\x12\x1a\n\x08position\x18\x01 \x01(\x05R\x08position\x12\x1f\n\x0bplayer_root\x18\x02 \x01(\x0cR\nplayerRoot\x12\x16\n\x06payout\x18\x03 \x01(\x03R\x06payout\"\x9f\x01\n\nBlindLevel\x12\x14\n\x05level\x18\x01 \x01(\x05R\x05level\x12\x1f\n\x0bsmall_blind\x18\x02 \x01(\x03R\nsmallBlind\x12\x1b\n\tbig_blind\x18\x03 \x01(\x03R\x08\x62igBlind\x12\x12\n\x04\x61nte\x18\x04 \x01(\x03R\x04\x61nte\x12)\n\x10\x64uration_minutes\x18\x05 \x01(\x05R\x0f\x64urationMinutes\"\xf1\x06\n\x0fTournamentState\x12#\n\rtournament_id\x18\x01 \x01(\tR\x0ctournamentId\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x38\n\x0cgame_variant\x18\x03 \x01(\x0e\x32\x15.examples.GameVariantR\x0bgameVariant\x12\x32\n\x06status\x18\x04 \x01(\x0e\x32\x1a.examples.TournamentStatusR\x06status\x12\x15\n\x06\x62uy_in\x18\x05 \x01(\x03R\x05\x62uyIn\x12%\n\x0estarting_stack\x18\x06 \x01(\x03R\rstartingStack\x12\x1f\n\x0bmax_players\x18\x07 \x01(\x05R\nmaxPlayers\x12\x1f\n\x0bmin_players\x18\x08 \x01(\x05R\nminPlayers\x12\x43\n\x0fscheduled_start\x18\t \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0escheduledStart\x12\x38\n\x0crebuy_config\x18\n \x01(\x0b\x32\x15.examples.RebuyConfigR\x0brebuyConfig\x12\x38\n\x0c\x61\x64\x64on_config\x18\x0b \x01(\x0b\x32\x15.examples.AddonConfigR\x0b\x61\x64\x64onConfig\x12=\n\x0f\x62lind_structure\x18\x0c \x03(\x0b\x32\x14.examples.BlindLevelR\x0e\x62lindStructure\x12#\n\rcurrent_level\x18\r \x01(\x05R\x0c\x63urrentLevel\x12_\n\x12registered_players\x18\x0e \x03(\x0b\x32\x30.examples.TournamentState.RegisteredPlayersEntryR\x11registeredPlayers\x12+\n\x11players_remaining\x18\x0f \x01(\x05R\x10playersRemaining\x12(\n\x10total_prize_pool\x18\x10 \x01(\x03R\x0etotalPrizePool\x1a\x62\n\x16RegisteredPlayersEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32\x1c.examples.PlayerRegistrationR\x05value:\x02\x38\x01*\xd2\x01\n\x10TournamentStatus\x12!\n\x1dTOURNAMENT_STATUS_UNSPECIFIED\x10\x00\x12\x16\n\x12TOURNAMENT_CREATED\x10\x01\x12 \n\x1cTOURNAMENT_REGISTRATION_OPEN\x10\x02\x12\x16\n\x12TOURNAMENT_RUNNING\x10\x03\x12\x15\n\x11TOURNAMENT_PAUSED\x10\x04\x12\x18\n\x14TOURNAMENT_COMPLETED\x10\x05\x12\x18\n\x14TOURNAMENT_CANCELLED\x10\x06\x42j\n\x14\x64\x65v.angzarr.examplesP\x01Z)github.com/angzarr/angzarr/proto/examples\xaa\x02\x10\x41ngzarr.Examples\xea\x02\x11\x41ngzarr::Examplesb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'examples.tournament_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\024dev.angzarr.examplesP\001Z)github.com/angzarr/angzarr/proto/examples\252\002\020Angzarr.Examples\352\002\021Angzarr::Examples'
  _globals['_TOURNAMENTSTATE_REGISTEREDPLAYERSENTRY']._loaded_options = None
  _globals['_TOURNAMENTSTATE_REGISTEREDPLAYERSENTRY']._serialized_options = b'8\001'
  _globals['_TOURNAMENTSTATUS']._serialized_start=5975
  _globals['_TOURNAMENTSTATUS']._serialized_end=6185
  _globals['_REBUYCONFIG']._serialized_start=101
  _globals['_REBUYCONFIG']._serialized_end=322
  _globals['_ADDONCONFIG']._serialized_start=325
  _globals['_ADDONCONFIG']._serialized_end=461
  _globals['_PLAYERREGISTRATION']._serialized_start=464
  _globals['_PLAYERREGISTRATION']._serialized_end=798
  _globals['_CREATETOURNAMENT']._serialized_start=801
  _globals['_CREATETOURNAMENT']._serialized_end=1273
  _globals['_OPENREGISTRATION']._serialized_start=1275
  _globals['_OPENREGISTRATION']._serialized_end=1293
  _globals['_CLOSEREGISTRATION']._serialized_start=1295
  _globals['_CLOSEREGISTRATION']._serialized_end=1314
  _globals['_ENROLLPLAYER']._serialized_start=1316
  _globals['_ENROLLPLAYER']._serialized_end=1402
  _globals['_UNREGISTERPLAYER']._serialized_start=1404
  _globals['_UNREGISTERPLAYER']._serialized_end=1455
  _globals['_STARTTOURNAMENT']._serialized_start=1457
  _globals['_STARTTOURNAMENT']._serialized_end=1474
  _globals['_PROCESSREBUY']._serialized_start=1476
  _globals['_PROCESSREBUY']._serialized_end=1562
  _globals['_PROCESSADDON']._serialized_start=1564
  _globals['_PROCESSADDON']._serialized_end=1650
  _globals['_ADVANCEBLINDLEVEL']._serialized_start=1652
  _globals['_ADVANCEBLINDLEVEL']._serialized_end=1671
  _globals['_ELIMINATEPLAYER']._serialized_start=1673
  _globals['_ELIMINATEPLAYER']._serialized_end=1752
  _globals['_PAUSETOURNAMENT']._serialized_start=1754
  _globals['_PAUSETOURNAMENT']._serialized_end=1795
  _globals['_RESUMETOURNAMENT']._serialized_start=1797
  _globals['_RESUMETOURNAMENT']._serialized_end=1815
  _globals['_TOURNAMENTCREATED']._serialized_start=1818
  _globals['_TOURNAMENTCREATED']._serialized_end=2350
  _globals['_REGISTRATIONOPENED']._serialized_start=2352
  _globals['_REGISTRATIONOPENED']._serialized_end=2429
  _globals['_REGISTRATIONCLOSED']._serialized_start=2431
  _globals['_REGISTRATIONCLOSED']._serialized_end=2557
  _globals['_TOURNAMENTPLAYERENROLLED']._serialized_start=2560
  _globals['_TOURNAMENTPLAYERENROLLED']._serialized_end=2834
  _globals['_TOURNAMENTENROLLMENTREJECTED']._serialized_start=2837
  _globals['_TOURNAMENTENROLLMENTREJECTED']._serialized_end=3024
  _globals['_PLAYERUNREGISTERED']._serialized_start=3027
  _globals['_PLAYERUNREGISTERED']._serialized_end=3186
  _globals['_TOURNAMENTSTARTED']._serialized_start=3189
  _globals['_TOURNAMENTSTARTED']._serialized_end=3385
  _globals['_REBUYPROCESSED']._serialized_start=3388
  _globals['_REBUYPROCESSED']._serialized_end=3636
  _globals['_REBUYDENIED']._serialized_start=3639
  _globals['_REBUYDENIED']._serialized_end=3805
  _globals['_ADDONPROCESSED']._serialized_start=3808
  _globals['_ADDONPROCESSED']._serialized_end=4023
  _globals['_BLINDLEVELADVANCED']._serialized_start=4026
  _globals['_BLINDLEVELADVANCED']._serialized_end=4211
  _globals['_PLAYERELIMINATED']._serialized_start=4214
  _globals['_PLAYERELIMINATED']._serialized_end=4424
  _globals['_TOURNAMENTPAUSED']._serialized_start=4426
  _globals['_TOURNAMENTPAUSED']._serialized_end=4525
  _globals['_TOURNAMENTRESUMED']._serialized_start=4527
  _globals['_TOURNAMENTRESUMED']._serialized_end=4605
  _globals['_TOURNAMENTCOMPLETED']._serialized_start=4608
  _globals['_TOURNAMENTCOMPLETED']._serialized_end=4821
  _globals['_TOURNAMENTRESULT']._serialized_start=4823
  _globals['_TOURNAMENTRESULT']._serialized_end=4926
  _globals['_BLINDLEVEL']._serialized_start=4929
  _globals['_BLINDLEVEL']._serialized_end=5088
  _globals['_TOURNAMENTSTATE']._serialized_start=5091
  _globals['_TOURNAMENTSTATE']._serialized_end=5972
  _globals['_TOURNAMENTSTATE_REGISTEREDPLAYERSENTRY']._serialized_start=5874
  _globals['_TOURNAMENTSTATE_REGISTEREDPLAYERSENTRY']._serialized_end=5972
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from angzarr_client.proto.angzarr import process_manager_pb2 as angzarr_dot_process__manager__pb2


class ProcessManagerServiceStub(object):
    """ProcessManagerService: stateful coordinator for long-running workflows across multiple aggregates.

    WARNING: Only use when saga + queries is insufficient. Consider:
    - Can a simple saga + destination queries solve this?
    - Is the "state" you want to track already derivable from existing aggregates?
    - Are you adding Process Manager because the workflow is genuinely complex?

    Process Manager is warranted when:
    - Workflow state is NOT derivable from aggregates (PM owns unique state)
    - You need to query workflow status independently ("show all pending fulfillments")
    - Timeout/scheduling logic is complex enough to merit its own aggregate
    - You must react to events from MULTIPLE domains (saga recommends single domain)

    Process Manager IS an aggregate with its own domain, events, and state.
    It reuses all aggregate infrastructure (EventStore, SnapshotStore, AggregateCoordinator).
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Prepare = channel.unary_unary(
                '/angzarr.ProcessManagerService/Prepare',
                request_serializer=angzarr_dot_process__manager__pb2.ProcessManagerPrepareRequest.SerializeToString,
                response_deserializer=angzarr_dot_process__manager__pb2.ProcessManagerPrepareResponse.FromString,
                _registered_method=True)
        self.Handle = channel.unary_unary(
                '/angzarr.ProcessManagerService/Handle',
                request_serializer=angzarr_dot_process__manager__pb2.ProcessManagerHandleRequest.SerializeToString,
                response_deserializer=angzarr_dot_process__manager__pb2.ProcessManagerHandleResponse.FromString,
                _registered_method=True)


class ProcessManagerServiceServicer(object):
    """ProcessManagerService: stateful coordinator for long-running workflows across multiple aggregates.

    WARNING: Only use when saga + queries is insufficient. Consider:
    - Can a simple saga + destination queries solve this?
    - Is the "state" you want to track already derivable from existing aggregates?
    - Are you adding Process Manager because the workflow is genuinely complex?

    Process Manager is warranted when:
    - Workflow state is NOT derivable from aggregates (PM owns unique state)
    - You need to query workflow status independently ("show all pending fulfillments")
    - Timeout/scheduling logic is complex enough to merit its own aggregate
    - You must react to events from MULTIPLE domains (saga recommends single domain)

    Process Manager IS an aggregate with its own domain, events, and state.
    It reuses all aggregate infrastructure (EventStore, SnapshotStore, AggregateCoordinator).
    """

    def Prepare(self, request, context):
        """Phase 1: Declare which additional destinations are needed beyond the trigger.
        PM automatically receives triggering event's domain state.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Handle(self, request, context):
        """Phase 2: Handle with trigger + process state + fetched destinations.
        Returns commands for other aggregates and events for the PM's own domain.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ProcessManagerServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Prepare': grpc.unary_unary_rpc_method_handler(
                    servicer.Prepare,
                    request_deserializer=angzarr_dot_process__manager__pb2.ProcessManagerPrepareRequest.FromString,
                    response_serializer=angzarr_dot_process__manager__pb2.ProcessManagerPrepareResponse.SerializeToString,
            ),
            'Handle': grpc.unary_unary_rpc_method_handler(
                    servicer.Handle,
                    request_deserializer=angzarr_dot_process__manager__pb2.ProcessManagerHandleRequest.FromString,
                    response_serializer=angzarr_dot_process__manager__pb2.ProcessManagerHandleResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'angzarr.ProcessManagerService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('angzarr.ProcessManagerService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class ProcessManagerService(object):
    """ProcessManagerService: stateful coordinator for long-running workflows across multiple aggregates.

    WARNING: Only use when saga + queries is insufficient. Consider:
    - Can a simple saga + destination queries solve this?
    - Is the "state" you want to track already derivable from existing aggregates?
    - Are you adding Process Manager because the workflow is genuinely complex?

    Process Manager is warranted when:
    - Workflow state is NOT derivable from aggregates (PM owns unique state)
    - You need to query workflow status independently ("show all pending fulfillments")
    - Timeout/scheduling logic is complex enough to merit its own aggregate
    - You must react to events from MULTIPLE domains (saga recommends single domain)

    Process Manager IS an aggregate with its own domain, events, and state.
    It reuses all aggregate infrastructure (EventStore, SnapshotStore, AggregateCoordinator).
    """

    @staticmethod
    def Prepare(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/angzarr.ProcessManagerService/Prepare',
            angzarr_dot_process__manager__pb2.ProcessManagerPrepareRequest.SerializeToString,
            angzarr_dot_process__manager__pb2.ProcessManagerPrepareResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Handle(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/angzarr.ProcessManagerService/Handle',
            angzarr_dot_process__manager__pb2.ProcessManagerHandleRequest.SerializeToString,
            angzarr_dot_process__manager__pb2.ProcessManagerHandleResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class ProcessManagerCoordinatorServiceStub(object):
    """ProcessManagerCoordinatorService: orchestrates PM execution
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Handle = channel.unary_unary(
                '/angzarr.ProcessManagerCoordinatorService/Handle',
                request_serializer=angzarr_dot_process__manager__pb2.ProcessManagerCoordinatorRequest.SerializeToString,
                response_deserializer=angzarr_dot_process__manager__pb2.ProcessManagerHandleResponse.FromString,
                _registered_method=True)
        self.HandleSpeculative = channel.unary_unary(
                '/angzarr.ProcessManagerCoordinatorService/HandleSpeculative',
                request_serializer=angzarr_dot_process__manager__pb2.SpeculatePmRequest.SerializeToString,
                response_deserializer=angzarr_dot_process__manager__pb2.ProcessManagerHandleResponse.FromString,
                _registered_method=True)


class ProcessManagerCoordinatorServiceServicer(object):
    """ProcessManagerCoordinatorService: orchestrates PM execution
    """

    def Handle(self, request, context):
        """Full orchestration with sync mode.
        Used by CASCADE mode to call PMs synchronously.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HandleSpeculative(self, request, context):
        """Speculative execution - returns commands and events without persisting
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ProcessManagerCoordinatorServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Handle': grpc.unary_unary_rpc_method_handler(
                    servicer.Handle,
                    request_deserializer=angzarr_dot_process__manager__pb2.ProcessManagerCoordinatorRequest.FromString,
                    response_serializer=angzarr_dot_process__manager__pb2.ProcessManagerHandleResponse.SerializeToString,
            ),
            'HandleSpeculative': grpc.unary_unary_rpc_method_handler(
                    servicer.HandleSpeculative,
                    request_deserializer=angzarr_dot_process__manager__pb2.SpeculatePmRequest.FromString,
                    response_serializer=angzarr_dot_process__manager__pb2.ProcessManagerHandleResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'angzarr.ProcessManagerCoordinatorService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('angzarr.ProcessManagerCoordinatorService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class ProcessManagerCoordinatorService(object):
    """ProcessManagerCoordinatorService: orchestrates PM execution
    """

    @staticmethod
    def Handle(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/angzarr.ProcessManagerCoordinatorService/Handle',
            angzarr_dot_process__manager__pb2.ProcessManagerCoordinatorRequest.SerializeToString,
            angzarr_dot_process__manager__pb2.ProcessManagerHandleResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def HandleSpeculative(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/angzarr.ProcessManagerCoordinatorService/HandleSpeculative',
            angzarr_dot_process__manager__pb2.SpeculatePmRequest.SerializeToString,
            angzarr_dot_process__manager__pb2.ProcessManagerHandleResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""Compensation flow helpers for saga/PM rejection handling.

This module provides helpers for implementing compensation logic in aggregates
and process managers when saga/PM commands are rejected.

Usage in Aggregate:
    from angzarr_client import Aggregate, handles
    from angzarr_client.compensation import (
        CompensationContext,
        delegate_to_framework,
        emit_compensation_events,
    )

    class OrderAggregate(Aggregate[OrderState]):
        def handle_revocation(self, notification):
            ctx = CompensationContext.from_notification(notification)

            # Option 1: Emit compensation events based on rejected command type
            if ctx.rejected_command_type == "type.googleapis.com/orders.FulfillOrder":
                event = OrderCancelled(
                    order_id=self.order_id,
                    reason=f"Fulfillment failed: {ctx.rejection_reason}",
                )
                self._apply_and_record(event)
                return emit_compensation_events(self.event_book())

            # Option 2: Delegate to framework
            return delegate_to_framework(
                reason=f"No custom compensation for {ctx.rejected_command_type}"
            )

Usage in ProcessManager:
    from angzarr_client.process_manager import ProcessManager, handles
    from angzarr_client.compensation import (
        CompensationContext,
        pm_delegate_to_framework,
        pm_emit_compensation_events,
    )

    class OrderWorkflowPM(ProcessManager[WorkflowState]):
        def handle_revocation(self, notification):
            ctx = CompensationContext.from_notification(notification)

            # Record failure in PM state
            event = WorkflowStepFailed(
                source_domain=ctx.source_domain,
                reason=ctx.rejection_reason,
            )
            self._apply_and_record(event)

            # Return PM events + framework response
            return pm_emit_compensation_events(
                process_events=self.process_events(),
                also_emit_system_event=True,
            )
"""

from dataclasses import dataclass

from .proto.angzarr import command_handler_pb2 as command_handler
from .proto.angzarr import types_pb2 as types


@dataclass
class CompensationContext:
    """Extracted context from a rejection Notification.

    Provides easy access to compensation-relevant fields. Source aggregate info
    is extracted from the rejected command's angzarr_deferred header.
    """

    source_event_sequence: int
    """Sequence of the event that triggered the saga/PM flow."""

    rejection_reason: str
    """Why the command was rejected."""

    rejected_command: types.CommandBook | None
    """The command that was rejected (if available)."""

    source_aggregate: types.Cover | None
    """Cover of the aggregate that triggered the flow."""

    @classmethod
    def from_notification(
        cls, notification: types.Notification
    ) -> "CompensationContext":
        """Extract compensation context from a Notification.

        Source aggregate info is extracted from the rejected command's
        angzarr_deferred header, which is always set by the framework
        for saga/PM-produced commands.

        Args:
            notification: The notification containing RejectionNotification payload.

        Returns:
            CompensationContext with extracted fields.
        """
        rejection = types.RejectionNotification()
        if notification.HasField("payload"):
            notification.payload.Unpack(rejection)

        # Extract source info from angzarr_deferred header
        source_aggregate = None
        source_event_sequence = 0

        if rejection.HasField("rejected_command"):
            cmd = rejection.rejected_command
            if cmd.pages:
                header = cmd.pages[0].header
                if header.HasField("angzarr_deferred"):
                    deferred = header.angzarr_deferred
                    source_event_sequence = deferred.source_seq
                    if deferred.HasField("source"):
                        source_aggregate = deferred.source

        return cls(
            source_event_sequence=source_event_sequence,
            rejection_reason=rejection.rejection_reason,
            rejected_command=(
                rejection.rejected_command
                if rejection.HasField("rejected_command")
                else None
            ),
            source_aggregate=source_aggregate,
        )

    @property
    def rejected_command_type(self) -> str | None:
        """Get the type URL of the rejected command, if available."""
        if self.rejected_command and self.rejected_command.pages:
            page = self.rejected_command.pages[0]
            if page.HasField("command"):
                return page.command.type_url
        return None

    @property
    def source_domain(self) -> str | None:
        """Get the domain of the source aggregate, if available."""
        if self.source_aggregate:
            return self.source_aggregate.domain
        return None

    @property
    def source_root(self) -> bytes | None:
        """Get the root UUID bytes of the source aggregate, if available."""
        if self.source_aggregate and self.source_aggregate.HasField("root"):
            return self.source_aggregate.root.value
        return None


# --- Aggregate helpers ---


def delegate_to_framework(
    reason: str,
    emit_system_event: bool = True,
    send_to_dead_letter: bool = False,
    escalate: bool = False,
    abort: bool = False,
) -> command_handler.BusinessResponse:
    """Create a response that delegates compensation to the framework.

    Use when the aggregate doesn't have custom compensation logic for a saga.
    The framework will emit a SagaCompensationFailed event to the fallback domain.

    Args:
        reason: Human-readable explanation for the delegation.
        emit_system_event: Emit SagaCompensationFailed to fallback domain.
        send_to_dead_letter: Move failed event to dead letter queue.
        escalate: Mark for operator intervention.
        abort: Stop the saga entirely without retry.

    Returns:
        BusinessResponse with revocation flags.
    """
    return command_handler.BusinessResponse(
        revocation=command_handler.RevocationResponse(
            emit_system_revocation=emit_system_event,
            send_to_dead_letter_queue=send_to_dead_letter,
            escalate=escalate,
            abort=abort,
            reason=reason,
        )
    )


def emit_compensation_events(
    event_book: types.EventBook,
) -> command_handler.BusinessResponse:
    """Create a response containing compensation events.

    Use when the aggregate emits events to record compensation.
    The framework will persist these events and NOT emit a system event.

    Args:
        event_book: EventBook containing compensation events.

    Returns:
        BusinessResponse with events.
    """
    return command_handler.BusinessResponse(events=event_book)


# --- Process Manager helpers ---


@dataclass
class RejectionHandlerResponse:
    """Response from rejection handlers.

    Can contain events (compensation), notification (upstream propagation), or both.
    """

    events: types.EventBook | None = None
    """Events to persist to own state (compensation)."""

    notification: types.Notification | None = None
    """Notification to forward upstream (rejection propagation)."""


def pm_delegate_to_framework(
    reason: str,
    emit_system_event: bool = True,
) -> tuple[None, command_handler.RevocationResponse]:
    """Create a PM response that delegates compensation to the framework.

    Use when the PM doesn't have custom compensation logic.

    Args:
        reason: Human-readable explanation for the delegation.
        emit_system_event: Emit SagaCompensationFailed to fallback domain.

    Returns:
        Tuple of (None, RevocationResponse) - no PM events, delegate to framework.
    """
    return None, command_handler.RevocationResponse(
        emit_system_revocation=emit_system_event,
        reason=reason,
    )


def pm_emit_compensation_events(
    process_events: types.EventBook,
    also_emit_system_event: bool = False,
    reason: str = "",
) -> tuple[types.EventBook, command_handler.RevocationResponse]:
    """Create a PM response containing compensation events.

    Use when the PM emits events to record the compensation in its state.

    Args:
        process_events: EventBook containing PM compensation events.
        also_emit_system_event: Also emit SagaCompensationFailed.
        reason: Reason for system event (if emitting).

    Returns:
        Tuple of (EventBook, RevocationResponse).
    """
    return process_events, command_handler.RevocationResponse(
        emit_system_revocation=also_emit_system_event,
        reason=reason,
    )



"""Query client step definitions."""

import uuid
from datetime import datetime
from unittest.mock import MagicMock

import pytest
from google.protobuf.any_pb2 import Any
from google.protobuf.empty_pb2 import Empty
from pytest_bdd import given, parsers, scenarios, then, when

from angzarr_client.proto.angzarr import types_pb2

# Link to feature file


@pytest.fixture
def query_context():
    """Test context for query client scenarios."""
    return {
        "client": None,
        "result": None,
        "error": None,
        "aggregates": {},
        "correlation_events": {},
        "service_available": True,
    }


def make_event_book(
    domain: str,
    root: str,
    event_count: int,
    start_seq: int = 0,
    snapshot_seq: int = None,
) -> types_pb2.EventBook:
    """Create an EventBook with specified events."""
    book = types_pb2.EventBook()
    book.cover.domain = domain
    try:
        root_uuid = uuid.UUID(root)
    except ValueError:
        root_uuid = uuid.uuid4()
    book.cover.root.value = root_uuid.bytes

    # Add snapshot if specified
    if snapshot_seq is not None:
        book.snapshot.sequence = snapshot_seq
        book.snapshot.state.Pack(Empty())

    # Add event pages
    for i in range(event_count):
        page = book.pages.add()
        page.header.sequence = start_seq + i
        page.event.Pack(Empty())

    return book


# ==========================================================================
# Background Steps
# ==========================================================================


@given("a QueryClient connected to the test backend")
def given_query_client(query_context):
    query_context["client"] = MagicMock()


# ==========================================================================
# Given Steps - Aggregates
# ==========================================================================


@given(parsers.parse('an aggregate "{domain}" with root "{root}"'))
def given_aggregate(query_context, domain, root):
    key = f"{domain}:{root}"
    query_context["aggregates"][key] = make_event_book(domain, root, 0)


@given(parsers.parse('an aggregate "{domain}" with root "{root}" has {count:d} events'))
def given_aggregate_with_events(query_context, domain, root, count, request):
    from tests.features.conftest import SHARED_EVENT_STORE

    key = f"{domain}:{root}"
    book = make_event_book(domain, root, count)
    query_context["aggregates"][key] = book
    # Also store in shared event store for cross-context access
    SHARED_EVENT_STORE[root] = book

    # Populate other contexts if available (for cross-context scenarios)
    for ctx_name in ["speculative_context", "domain_client_context"]:
        try:
            ctx = request.getfixturevalue(ctx_name)
            ctx["event_book"] = book
            if ctx_name == "speculative_context":
                ctx["base_event_count"] = count
        except Exception:
            pass


@given(
    parsers.parse(
        'an aggregate "{domain}" with root "{root}" has event "{event_type}" with data "{data}"'
    )
)
def given_aggregate_with_specific_event(query_context, domain, root, event_type, data):
    key = f"{domain}:{root}"
    book = types_pb2.EventBook()
    book.cover.domain = domain
    try:
        root_uuid = uuid.UUID(root)
    except ValueError:
        root_uuid = uuid.uuid4()
    book.cover.root.value = root_uuid.bytes

    page = book.pages.add()
    page.header.sequence = 0
    # Store event type in type_url
    page.event.type_url = f"type.googleapis.com/{event_type}"
    page.event.value = data.encode()

    query_context["aggregates"][key] = book


@given(
    parsers.parse(
        'an aggregate "{domain}" with root "{root}" has events at known timestamps'
    )
)
def given_aggregate_with_timestamps(query_context, domain, root):
    key = f"{domain}:{root}"
    query_context["aggregates"][key] = make_event_book(domain, root, 5)


@given(
    parsers.parse('an aggregate "{domain}" with root "{root}" in edition "{edition}"')
)
def given_aggregate_in_edition(query_context, domain, root, edition):
    key = f"{domain}:{root}:{edition}"
    query_context["aggregates"][key] = make_event_book(domain, root, 3)


@given(
    parsers.parse(
        'an aggregate "{domain}" with root "{root}" has {count:d} events in main'
    )
)
def given_aggregate_in_main(query_context, domain, root, count):
    key = f"{domain}:{root}"
    query_context["aggregates"][key] = make_event_book(domain, root, count)


@given(
    parsers.parse(
        'an aggregate "{domain}" with root "{root}" has {count:d} events in edition "{edition}"'
    )
)
def given_aggregate_in_edition_with_count(query_context, domain, root, count, edition):
    key = f"{domain}:{root}:{edition}"
    query_context["aggregates"][key] = make_event_book(domain, root, count)


@given(
    parsers.parse(
        'an aggregate "{domain}" with root "{root}" has a snapshot at sequence {snap:d} and {total:d} events'
    )
)
def given_aggregate_with_snapshot(query_context, domain, root, snap, total):
    key = f"{domain}:{root}"
    query_context["aggregates"][key] = make_event_book(
        domain, root, total, snapshot_seq=snap
    )


@given(parsers.parse('events with correlation ID "{cid}" exist in multiple aggregates'))
def given_correlated_events(query_context, cid):
    query_context["correlation_events"][cid] = [
        make_event_book("orders", "order-1", 2),
        make_event_book("inventory", "inv-1", 1),
    ]


@given("the query service is unavailable")
def given_service_unavailable(query_context):
    query_context["service_available"] = False


# ==========================================================================
# When Steps
# ==========================================================================


@when(parsers.parse('I query events for "{domain}" root "{root}"'))
def when_query_events(query_context, domain, root):
    if not query_context["service_available"]:
        query_context["error"] = ConnectionError("Service unavailable")
        return

    key = f"{domain}:{root}"
    if key in query_context["aggregates"]:
        query_context["result"] = query_context["aggregates"][key]
    else:
        query_context["result"] = make_event_book(domain, root, 0)


@when(
    parsers.parse('I query events for "{domain}" root "{root}" from sequence {start:d}')
)
def when_query_events_from_sequence(query_context, domain, root, start):
    key = f"{domain}:{root}"
    if key in query_context["aggregates"]:
        full_book = query_context["aggregates"][key]
        # Filter pages from start sequence
        result = types_pb2.EventBook()
        result.cover.CopyFrom(full_book.cover)
        for page in full_book.pages:
            if page.header.sequence >= start:
                new_page = result.pages.add()
                new_page.CopyFrom(page)
        query_context["result"] = result
    else:
        query_context["result"] = make_event_book(domain, root, 0)


@when(
    parsers.parse(
        'I query events for "{domain}" root "{root}" from sequence {start:d} to {end:d}'
    )
)
def when_query_events_range(query_context, domain, root, start, end):
    key = f"{domain}:{root}"
    if key in query_context["aggregates"]:
        full_book = query_context["aggregates"][key]
        result = types_pb2.EventBook()
        result.cover.CopyFrom(full_book.cover)
        for page in full_book.pages:
            if start <= page.header.sequence < end:
                new_page = result.pages.add()
                new_page.CopyFrom(page)
        query_context["result"] = result
    else:
        query_context["result"] = make_event_book(domain, root, 0)


@when(
    parsers.parse('I query events for "{domain}" root "{root}" as of sequence {seq:d}')
)
def when_query_events_as_of_sequence(query_context, domain, root, seq):
    key = f"{domain}:{root}"
    if key in query_context["aggregates"]:
        full_book = query_context["aggregates"][key]
        result = types_pb2.EventBook()
        result.cover.CopyFrom(full_book.cover)
        for page in full_book.pages:
            if page.header.sequence <= seq:
                new_page = result.pages.add()
                new_page.CopyFrom(page)
        query_context["result"] = result
    else:
        query_context["result"] = make_event_book(domain, root, 0)


@when(
    parsers.parse(
        'I query events for "{domain}" root "{root}" as of time "{timestamp}"'
    )
)
def when_query_events_as_of_time(query_context, domain, root, timestamp):
    key = f"{domain}:{root}"
    if key in query_context["aggregates"]:
        # For testing, return all events (timestamp filtering is simulated)
        query_context["result"] = query_context["aggregates"][key]
    else:
        query_context["result"] = make_event_book(domain, root, 0)


@when(
    parsers.parse('I query events for "{domain}" root "{root}" in edition "{edition}"')
)
def when_query_events_in_edition(query_context, domain, root, edition):
    key = f"{domain}:{root}:{edition}"
    if key in query_context["aggregates"]:
        query_context["result"] = query_context["aggregates"][key]
    else:
        query_context["result"] = make_event_book(domain, root, 0)


@when(parsers.parse('I query events by correlation ID "{cid}"'))
def when_query_by_correlation_id(query_context, cid):
    if cid in query_context["correlation_events"]:
        # Combine events from all correlated aggregates
        result = types_pb2.EventBook()
        for book in query_context["correlation_events"][cid]:
            for page in book.pages:
                new_page = result.pages.add()
                new_page.CopyFrom(page)
        query_context["result"] = result
    else:
        query_context["result"] = types_pb2.EventBook()


@when("I query events with empty domain")
def when_query_empty_domain(query_context):
    query_context["error"] = ValueError("Invalid argument: empty domain")


@when("I attempt to query events")
def when_attempt_query(query_context):
    if not query_context["service_available"]:
        query_context["error"] = ConnectionError("Connection error")


# ==========================================================================
# Then Steps
# ==========================================================================


@then(parsers.parse("I should receive an EventBook with {count:d} events"))
def then_receive_event_book_with_count(query_context, count):
    result = query_context.get("result")
    assert result is not None, "Should have a result"
    assert (
        len(result.pages) == count
    ), f"Expected {count} events, got {len(result.pages)}"


@then(parsers.parse("the next_sequence should be {seq:d}"))
def then_next_sequence_is(query_context, seq):
    result = query_context.get("result")
    assert result is not None
    # next_sequence = number of events
    assert len(result.pages) == seq


@then(parsers.parse("events should be in sequence order {start:d} to {end:d}"))
def then_events_in_sequence_order(query_context, start, end):
    result = query_context.get("result")
    assert result is not None
    for i, page in enumerate(result.pages):
        assert (
            page.header.sequence == start + i
        ), f"Expected sequence {start + i}, got {page.header.sequence}"


@then(parsers.parse('the first event should have type "{event_type}"'))
def then_first_event_has_type(query_context, event_type):
    result = query_context.get("result")
    assert result is not None
    assert len(result.pages) > 0
    assert event_type in result.pages[0].event.type_url


@then(parsers.parse('the first event should have payload "{payload}"'))
def then_first_event_has_payload(query_context, payload):
    result = query_context.get("result")
    assert result is not None
    assert len(result.pages) > 0
    assert result.pages[0].event.value.decode() == payload


@then(parsers.parse("the first event should have sequence {seq:d}"))
def then_first_event_has_sequence(query_context, seq):
    result = query_context.get("result")
    assert result is not None
    assert len(result.pages) > 0
    assert result.pages[0].header.sequence == seq


@then(parsers.parse("the last event should have sequence {seq:d}"))
def then_last_event_has_sequence(query_context, seq):
    result = query_context.get("result")
    assert result is not None
    assert len(result.pages) > 0
    assert result.pages[-1].header.sequence == seq


@then("I should receive events up to that timestamp")
def then_receive_events_up_to_timestamp(query_context):
    result = query_context.get("result")
    assert result is not None


@then("I should receive events from that edition only")
def then_receive_events_from_edition(query_context):
    result = query_context.get("result")
    assert result is not None


@then("I should receive events from all correlated aggregates")
def then_receive_correlated_events(query_context):
    result = query_context.get("result")
    assert result is not None
    assert len(result.pages) > 0


@then("I should receive no events")
def then_receive_no_events(query_context):
    result = query_context.get("result")
    assert result is not None
    assert len(result.pages) == 0


@then("the EventBook should include the snapshot")
def then_event_book_includes_snapshot(query_context):
    result = query_context.get("result")
    assert result is not None
    assert result.HasField("snapshot")


@then(parsers.parse("the returned snapshot should be at sequence {seq:d}"))
def then_snapshot_at_sequence(query_context, seq):
    result = query_context.get("result")
    assert result is not None
    assert result.snapshot.sequence == seq


@then("the operation should fail with invalid argument error")
def then_fail_with_invalid_argument(query_context):
    error = query_context.get("error")
    assert error is not None
    assert "invalid" in str(error).lower()


@then("the operation should fail with connection error")
def then_fail_with_connection_error(query_context):
    error = query_context.get("error")
    assert error is not None
    assert isinstance(error, ConnectionError) or "connection" in str(error).lower()
